#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
"""evohomeasync provides an async client for the *original* Evohome API."""
from __future__ import annotations

import asyncio
from datetime import datetime as dt
from http import HTTPMethod, HTTPStatus
import logging
from typing import TYPE_CHECKING, Any, Callable, NoReturn

import aiohttp

from .exceptions import (
    AuthenticationFailed,
    DeprecationError,
    InvalidSchema,
    RateLimitExceeded,
    RequestFailed,
)


if TYPE_CHECKING:
    from .schema import (
        _DeviceIdT,
        _EvoDictT,
        _EvoListT,
        _LocationIdT,
        _SystemModeT,
        _TaskIdT,
        _ZoneIdT,
    )


URL_HOST = "https://tccna.honeywell.com"


_LOGGER = logging.getLogger(__name__)


class EvohomeClientDeprecated:  # NOTE: incl. _wait_for_put_task()
    """Deprecated attributes and methods removed from the evohome-client namespace."""

    hostname: str

    async def _wait_for_put_task(self, response: aiohttp.ClientResponse) -> None:
        """This functionality is deprecated, but remains here as documentation."""

        async def get_task_status(task_id: _TaskIdT) -> str:
            await self._populate_full_data()

            url = f"/commTasks?commTaskId={task_id}"
            response = await self._do_request(HTTPMethod.GET, url)

            return dict(await response.json())["state"]

        task_id: _TaskIdT

        assert response.method == HTTPMethod.PUT

        ret = await response.json()
        task_id = ret[0]["id"] if isinstance(ret, list) else ret["id"]

        # FIXME: could wait forvever?
        while await get_task_status(task_id) != "Succeeded":
            await asyncio.sleep(1)

    # Note deprecated, just a placeholder for self.get_task_status()
    async def _do_request(self, *args, **kwargs) -> aiohttp.ClientResponse:
        raise NotImplementedError

    # Note deprecated, just a placeholder for self.get_task_status()
    async def _populate_full_data(self, *args, **kwargs) -> None:
        raise NotImplementedError

    async def get_system_modes(self, *args, **kwargs) -> NoReturn:
        raise DeprecationError(
            "EvohomeClient.get_modes() is deprecated, "
            "use .get_system_modes() or .get_zone_modes()"
        )

    async def set_status_away(self, *args, **kwargs) -> NoReturn:
        raise DeprecationError(
            "EvohomeClient.set_status_away() is deprecated, use .set_mode_away()"
        )

    async def set_status_custom(self, *args, **kwargs) -> NoReturn:
        raise DeprecationError(
            "EvohomeClient.set_status_custom() is deprecated, use .set_mode_custom()"
        )

    async def set_status_dayoff(self, *args, **kwargs) -> NoReturn:
        raise DeprecationError(
            "EvohomeClient.set_status_dayoff() is deprecated, use .set_mode_dayoff()"
        )

    async def set_status_eco(self, *args, **kwargs) -> NoReturn:
        raise DeprecationError(
            "EvohomeClient.set_status_eco() is deprecated, use .set_mode_eco()"
        )

    async def set_status_heatingoff(self, *args, **kwargs) -> NoReturn:
        raise DeprecationError(
            "EvohomeClient.set_status_heatingoff() is deprecated, use .set_mode_heatingoff()"
        )

    async def set_status_normal(self, *args, **kwargs) -> NoReturn:
        raise DeprecationError(
            "EvohomeClient.set_status_normal() is deprecated, use .set_mode_auto()"
        )


class EvohomeClient(EvohomeClientDeprecated):
    """Provide a client to access the Honeywell TCC API (assumes a single TCS)."""

    def __init__(
        self,
        username: str,
        password: str,
        /,
        *,
        user_data: dict[str, str] | None = None,
        **kwargs,
    ) -> None:
        """Construct the v1 EvohomeClient object.

        If user_data is given then this will be used to try and reduce the number of
        calls to the authentication service which is known to be rate limited.
        """
        if kwargs.get("debug"):
            _LOGGER.setLevel(logging.DEBUG)
            _LOGGER.debug("Debug mode is explicitly enabled.")

        self.username = username
        self.password = password

        self.user_data: dict[str, str] | None = user_data
        self.hostname: str = kwargs.get("hostname", URL_HOST)

        self._session = kwargs.get("session") or aiohttp.ClientSession(
            timeout=aiohttp.ClientTimeout(total=30)
        )

        self.full_data: _EvoDictT = None  # type: ignore[assignment]
        self.location_id: _LocationIdT = None  # type: ignore[assignment]

        self.devices: _EvoDictT = {}
        self.named_devices: _EvoDictT = {}

        self.postdata: dict[str, Any] = {}
        self.headers: dict[str, Any] = {}

    async def _do_request(
        self,
        method: HTTPMethod,
        url: str,
        /,
        *,
        data: dict | None = None,
        retry: bool = True,
    ) -> aiohttp.ClientResponse:
        """Perform an HTTP request, with optional retry (usu. for authentication)."""

        url = self.hostname + "/WebAPI/api" + url

        if method == HTTPMethod.GET:
            func = self._session.get
        elif method == HTTPMethod.PUT:
            func = self._session.put
        elif method == HTTPMethod.POST:
            func = self._session.post

        try:
            response = await self._do_request_base(func, url, data=data, retry=retry)

        except aiohttp.ClientError as exc:
            if method == HTTPMethod.POST:  # using response will cause UnboundLocalError
                raise AuthenticationFailed(str(exc))
            raise RequestFailed(str(exc))

        try:
            response.raise_for_status()

        except aiohttp.ClientResponseError as exc:
            if response.method == HTTPMethod.POST:  # POST only used when authenticating
                raise AuthenticationFailed(str(exc), status=exc.status)
            if response.status != HTTPStatus.TOO_MANY_REQUESTS:
                raise RateLimitExceeded(str(exc), status=exc.status)
            raise RequestFailed(str(exc), status=exc.status)

        except aiohttp.ClientError as exc:
            if response.method == HTTPMethod.POST:  # POST only used when authenticating
                raise AuthenticationFailed(str(exc))
            raise RequestFailed(str(exc))

        return response

    async def _do_request_base(
        self,
        func: Callable,
        url: str,
        /,
        *,
        data: dict | None = None,
        retry: bool = True,
    ) -> aiohttp.ClientResponse:
        """Perform an HTTP request, with optional retry (usu. for authentication)."""

        response: aiohttp.ClientResponse

        async with func(url, json=data, headers=self.headers) as response:  # NB: json=
            response_text = await response.text()  # why cant I move this below the if?

            # if 401/unauthorized, may need to refresh sessionId
            if response.status != HTTPStatus.UNAUTHORIZED or not retry:
                return response

            # TODO: use response.content_type to determine whether to use .json()
            if "code" not in response_text:  # don't use .json() yet: may be plain text
                return response

            response_json = await response.json()
            if response_json[0]["code"] != "Unauthorized":
                return response

            _LOGGER.debug("Session expired/invalid, re-authenticating...")
            self.user_data = None  # Get a fresh (= None) sessionId
            await self._populate_user_data()

            session_id = self.user_data["sessionId"]  # type: ignore[index]
            self.headers["sessionId"] = session_id
            _LOGGER.debug(f"... Success: sessionId = {session_id}")

            # NOTE: this is a recursive call, used only after re-authenticating
            response = await self._do_request_base(func, url, data=data, retry=False)

        return response

    async def _populate_user_data(self) -> _EvoDictT:
        """Retrieve all the user data from the web."""

        if self.user_data is None:
            self.postdata = {
                "Username": self.username,
                "Password": self.password,
                "ApplicationId": "91db1612-73fd-4500-91b2-e63b069b185c",
            }
            self.headers = {"content-type": "application/json"}

            url = "/session"
            response = await self._do_request(
                HTTPMethod.POST, url, data=self.postdata, retry=False
            )

            self.user_data = await response.json()

        assert isinstance(self.user_data, dict)  # mypy
        return self.user_data

    async def _populate_full_data(self, force_refresh: bool = False) -> None:
        """Retrieve all the system data from the web."""

        if self.full_data is None or force_refresh:
            await self._populate_user_data()

            self.headers["sessionId"] = self.user_data["sessionId"]  # type: ignore[index]
            user_id = self.user_data["userInfo"]["userID"]  # type: ignore[index]

            url = f"/locations?userId={user_id}&allData=True"
            response = await self._do_request(HTTPMethod.GET, url, data=self.postdata)

            self.full_data = list(await response.json())[0]
            self.location_id = self.full_data["locationID"]

            self.devices = {}
            self.named_devices = {}

            for device in self.full_data["devices"]:
                self.devices[device["deviceID"]] = device
                self.named_devices[device["name"]] = device

    async def temperatures(self, force_refresh: bool = False) -> _EvoListT:
        """Retrieve the current details for each zone (incl. DHW)."""

        set_point: float
        status: str

        await self._populate_full_data(force_refresh=force_refresh)

        result = []

        try:
            for device in self.full_data["devices"]:
                temp = float(device["thermostat"]["indoorTemperature"])
                values = device["thermostat"]["changeableValues"]

                if "heatSetpoint" in values:
                    set_point = float(values["heatSetpoint"]["value"])
                    status = values["heatSetpoint"]["status"]
                else:
                    set_point = 0
                    status = values["status"]

                result.append(
                    {
                        "thermostat": device["thermostatModelType"],
                        "id": device["deviceID"],
                        "name": device["name"],
                        "temp": None if temp == 128 else temp,
                        "setpoint": set_point,
                        "status": status,
                        "mode": values["mode"],
                    }
                )

        # harden code against unexpected schema (JSON structure)
        except (LookupError, TypeError, ValueError) as exc:
            raise InvalidSchema(str(exc))
        return result

    async def get_system_modes(self) -> NoReturn:
        """Return the set of modes the system can be assigned."""
        raise NotImplementedError

    async def _set_system_mode(
        self, status: _SystemModeT, until: dt | None = None
    ) -> None:
        """Set the system mode."""

        await self._populate_full_data()

        data: dict[str, str | None] = {"QuickAction": status}
        if until is None:
            data |= {"QuickActionNextTime": None}
        else:
            data |= {"QuickActionNextTime": until.strftime("%Y-%m-%dT%H:%M:%SZ")}

        url = f"/evoTouchSystems?locationId={self.location_id}"
        await self._do_request(HTTPMethod.PUT, url, data=data)

    async def set_mode_auto(self) -> None:
        """Set the system to normal operation."""
        await self._set_system_mode("Auto")

    async def set_mode_away(self, until: dt | None = None) -> None:
        """Set the system to the away mode."""
        await self._set_system_mode("Away", until)

    async def set_mode_custom(self, until: dt | None = None) -> None:
        """Set the system to the custom programme."""
        await self._set_system_mode("Custom", until)

    async def set_mode_dayoff(self, until: dt | None = None) -> None:
        """Set the system to the day off mode."""
        await self._set_system_mode("DayOff", until)

    async def set_mode_eco(self, until: dt | None = None) -> None:
        """Set the system to the eco mode."""
        await self._set_system_mode("AutoWithEco", until)

    async def set_mode_heatingoff(self, until: dt | None = None) -> None:
        """Set the system to the heating off mode."""
        await self._set_system_mode("HeatingOff", until)

    async def get_zone_modes(self, zone) -> list[str]:
        """Return the set of modes the zone can be assigned."""

        await self._populate_full_data()

        device = self._get_device(zone)
        return device["thermostat"]["allowedModes"]

    def _get_device(self, zone: str) -> _EvoDictT:
        """"""

        if isinstance(zone, str):
            return self.named_devices[zone]
        return self.devices[zone]  # TODO: no need for str.isnumeric() check?

    def _get_device_id(self, device_id: _DeviceIdT) -> _DeviceIdT:
        """"""

        device = self._get_device(device_id)
        return device["deviceID"]

    async def _set_heat_setpoint(self, zone, data: _EvoDictT) -> None:
        """"""

        await self._populate_full_data()

        zone_id: _ZoneIdT = self._get_device_id(zone)

        url = f"/devices/{zone_id}/thermostat/changeableValues/heatSetpoint"
        await self._do_request(HTTPMethod.PUT, url, data=data)

    async def set_temperature(self, zone, temperature, until: dt | None = None) -> None:
        """Set the temperature of the given zone."""

        if until is None:
            data = {"Value": temperature, "Status": "Hold", "NextTime": None}
        else:
            data = {
                "Value": temperature,
                "Status": "Temporary",
                "NextTime": until.strftime("%Y-%m-%dT%H:%M:%SZ"),
            }

        await self._set_heat_setpoint(zone, data)

    async def cancel_temp_override(self, zone) -> None:
        """Remove an existing temperature override."""

        data = {"Value": None, "Status": "Scheduled", "NextTime": None}
        await self._set_heat_setpoint(zone, data)

    def _get_dhw_zone(self) -> str | None:
        for device in self.full_data["devices"]:
            if device["thermostatModelType"] == "DOMESTIC_HOT_WATER":
                return device["deviceID"]
        return None

    async def _set_dhw(
        self,
        status: str = "Scheduled",
        mode: str | None = None,
        next_time: str | None = None,
    ) -> None:
        """Set DHW to On, Off or Auto, either indefinitely, or until a set time."""

        await self._populate_full_data()

        dhw_id = self._get_dhw_zone()
        if dhw_id is None:
            raise InvalidSchema("No DHW zone reported from API")

        data = {
            "Status": status,
            "Mode": mode,
            "NextTime": next_time,
            "SpecialModes": None,
            "HeatSetpoint": None,
            "CoolSetpoint": None,
        }

        url = f"/devices/{dhw_id}/thermostat/changeableValues"
        await self._do_request(HTTPMethod.PUT, url, data=data)

    async def set_dhw_on(self, until: dt | None = None) -> None:
        """Set DHW to on, either indefinitely, or until a specified time.

        When On, the DHW controller will work to keep its target temperature at/above
        its target temperature.  After the specified time, it will revert to its
        scheduled behaviour.
        """

        time_until = None if until is None else until.strftime("%Y-%m-%dT%H:%M:%SZ")

        await self._set_dhw(status="Hold", mode="DHWOn", next_time=time_until)

    async def set_dhw_off(self, until: dt | None = None) -> None:
        """Set DHW to on, either indefinitely, or until a specified time.

        When Off, the DHW controller will ignore its target temperature. After the
        specified time, it will revert to its scheduled behaviour.
        """

        time_until = None if until is None else until.strftime("%Y-%m-%dT%H:%M:%SZ")

        await self._set_dhw(status="Hold", mode="DHWOff", next_time=time_until)

    async def set_dhw_auto(self) -> None:
        """Set DHW to On or Off, according to its schedule."""
        await self._set_dhw(status="Scheduled")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
"""Hacked aiohttp to provide a mocked server."""
from __future__ import annotations

import asyncio
from enum import EnumCheck, StrEnum, verify
from types import TracebackType
from typing import TYPE_CHECKING, Any, Final, Self, Type

import json

from .vendor import MockedServer


if TYPE_CHECKING:
    from .const import _bodyT, _methodT, _statusT, _urlT


_DEFAULT_LIMIT = 2**16  # 64 KiB


@verify(EnumCheck.UNIQUE)
class hdrs(StrEnum):  # from aiohttp
    METH_GET: Final[str] = "GET"
    METH_POST: Final[str] = "POST"
    METH_PUT: Final[str] = "PUT"


class StreamReader(asyncio.StreamReader):
    _buffer = bytearray()

    def __init__(self, limit=_DEFAULT_LIMIT, loop=None):
        pass

    async def read(self, n: int = -1) -> bytes:
        """Read up to `n` bytes from the stream."""
        raise NotImplementedError

    async def readline(self) -> bytes:
        """Read chunk of data from the stream until newline (b'\n') is found."""
        return await self.readuntil(b"\n")

    async def readexactly(self, n: int) -> bytes:
        """Read exactly `n` bytes."""
        raise NotImplementedError

    async def readuntil(self, separator: bytes = b"\n") -> bytes:
        """Read data from the stream until ``separator`` is found."""
        raise NotImplementedError

    async def __aenter__(self, *args, **kwargs) -> Self:
        return self

    async def __aexit__(
        self,
        exc_type: Type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        pass


class ClientError(Exception):
    """Base class for client connection errors."""


class ClientResponseError(ClientError):
    """Base class for exceptions that occur after getting a response."""

    def __init__(self, msg, /, *, status: int | None = None, **kwargs) -> None:
        super().__init__(msg)
        self.status = status


class ClientTimeout:
    """"""

    def __init__(self, /, *, total: float | None = None, **kwargs) -> None:
        self.total: float = total or 30


class ClientSession:
    """First-class interface for making HTTP requests."""

    def __init__(self, /, *, timeout: ClientTimeout | None = None, **kwargs) -> None:
        self._timeout = timeout or ClientTimeout()

        # this is required, so no .get()
        self._mocked_server: MockedServer = kwargs["mocked_server"]

    def get(self, url, /, headers: str | None = None):
        return ClientResponse(hdrs.METH_GET, url, session=self)  # type: ignore[arg-type]

    def put(
        self, url, /, *, data: Any = None, json: Any = None, headers: str | None = None
    ):
        return ClientResponse(hdrs.METH_PUT, url, data=data or json, session=self)  # type: ignore[arg-type]

    def post(self, url, /, *, data: Any = None, headers: str | None = None):
        return ClientResponse(hdrs.METH_POST, url, data=data, session=self)  # type: ignore[arg-type]

    async def close(self) -> None:
        pass


class ClientResponse:
    """"""

    charset: str = "utf-8"

    def __init__(
        self,
        method: _methodT,
        url: _urlT,
        /,
        *,
        data: str | None = None,
        json: dict | None = None,
        session: ClientSession | None = None,
        **kwargs,
    ) -> None:
        self.method = method
        self.url = url
        self.session = session

        self._mocked_server = self.session._mocked_server

        self.status: _statusT = None
        self._body: _bodyT | None = None

        # self.content = StreamReader(
        #     self._mocked_server.request(method, url, data=data or json)
        # )
        self._body = self._mocked_server.request(method, url, data=data or json)
        self.status = self._mocked_server.status

    def raise_for_status(self) -> None:
        if self.status >= 300:
            raise ClientResponseError(
                f"{self.method} {self.url}: {self.status}", status=self.status
            )

    @property
    def content_length(self) -> int | None:
        if self._body is None:
            return None
        if self.content_type == "text/plain":
            return len(self._body)
        return len(str(self._body))

    @property
    def content_type(self) -> str | None:
        """Return the Content-Type header of the response."""

        # if isinstance(self._body, bytes):
        #     return "application/octet-stream"
        if isinstance(self._body, (dict, list)):
            return "application/json"
        if not isinstance(self._body, str):
            raise NotImplementedError
        if self._body.strip().startswith("<html>"):
            return "text/html"
        return "text/plain"

    async def text(self, /, **kwargs) -> str:  # assumes is JSON or plaintext
        """Return the response body as text."""
        if self.content_type == "application/json":
            return json.dumps(self._body)
        if self.content_type in ("text/html", "text/plain"):
            return self._body
        raise NotImplementedError

    async def json(self, /, **kwargs) -> dict | list:  # assumes is JSON or plaintext
        """Return the response body as json (a dict)."""
        if self.content_type == "application/json":
            return self._body
        assert not isinstance(self._body, (dict, list))  # mypy
        return json.loads(self._body)

    async def __aenter__(self, *args, **kwargs) -> Self:
        return self

    async def __aexit__(
        self,
        exc_type: Type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
"""Provides handling of TCC locations."""
from __future__ import annotations

from typing import TYPE_CHECKING, Final, NoReturn

from .exceptions import DeprecationError, InvalidSchema
from .gateway import Gateway
from .schema import SCH_LOCN_STATUS
from .schema.const import (
    SZ_COUNTRY,
    SZ_DHW,
    SZ_GATEWAY_ID,
    SZ_GATEWAYS,
    SZ_LOCATION,
    SZ_LOCATION_ID,
    SZ_LOCATION_INFO,
    SZ_LOCATION_OWNER,
    SZ_LOCATION_TYPE,
    SZ_NAME,
    SZ_SYSTEM_ID,
    SZ_TEMPERATURE_CONTROL_SYSTEMS,
    SZ_TIME_ZONE,
    SZ_USE_DAYLIGHT_SAVE_SWITCHING,
    SZ_ZONE_ID,
    SZ_ZONES,
)


if TYPE_CHECKING:
    import logging

    from . import Broker, ControlSystem, EvohomeClient
    from .schema import _EvoDictT, _LocationIdT


class _LocationDeprecated:
    """Deprecated attributes and methods removed from the evohome-client namespace."""

    async def status(self, *args, **kwargs) -> NoReturn:
        raise DeprecationError("Location.status() is deprecated, use .refresh_status()")


class Location(_LocationDeprecated):
    """Instance of an account's location."""

    STATUS_SCHEMA: Final = SCH_LOCN_STATUS
    TYPE: Final[str] = SZ_LOCATION

    def __init__(self, client: EvohomeClient, config: _EvoDictT) -> None:
        self.client = client

        self._broker: Broker = client._broker
        self._logger: logging.Logger = client._logger

        self._config: Final[_EvoDictT] = config[SZ_LOCATION_INFO]

        try:
            assert self.locationId, "Invalid config dict"
        except AssertionError as exc:
            raise InvalidSchema(str(exc))
        self._id = self.locationId

        self._gateways: list[Gateway] = []
        self.gateways: dict[str, Gateway] = {}  # gwy by id

        for gwy_config in config[SZ_GATEWAYS]:
            gwy = Gateway(self, gwy_config)

            self._gateways.append(gwy)
            self.gateways[gwy.gatewayId] = gwy

    def __str__(self) -> str:
        return f"{self._id} ({self.TYPE})"

    @property
    def country(self) -> str:
        return self._config[SZ_COUNTRY]

    @property
    def locationOwner(self) -> _EvoDictT:
        return self._config[SZ_LOCATION_OWNER]

    @property
    def locationId(self) -> _LocationIdT:
        return self._config[SZ_LOCATION_ID]

    @property
    def locationType(self) -> str:
        return self._config[SZ_LOCATION_TYPE]

    @property
    def name(self) -> str:
        return self._config[SZ_NAME]

    @property
    def timeZone(self) -> _EvoDictT:
        return self._config[SZ_TIME_ZONE]

    @property
    def useDaylightSaveSwitching(self) -> bool:
        return self._config[SZ_USE_DAYLIGHT_SAVE_SWITCHING]

    async def refresh_status(self) -> _EvoDictT:
        """Update the Location with its latest status (also returns the status)."""

        status: _EvoDictT = await self._broker.get(
            f"{self.TYPE}/{self._id}/status?includeTemperatureControlSystems=True",
            schema=self.STATUS_SCHEMA,
        )  # type: ignore[assignment]

        self._update_status(status)
        return status

    def _update_status(self, loc_status: _EvoDictT) -> None:
        tcs: ControlSystem
        gwy_status: _EvoDictT
        dhw_status: _EvoDictT
        tcs_status: _EvoDictT
        zon_status: _EvoDictT

        for gwy_status in loc_status[SZ_GATEWAYS]:
            gwy = self.gateways[gwy_status[SZ_GATEWAY_ID]]
            gwy._update_status(gwy_status)

            for tcs_status in gwy_status[SZ_TEMPERATURE_CONTROL_SYSTEMS]:
                tcs = gwy.control_systems[tcs_status[SZ_SYSTEM_ID]]
                tcs._update_status(tcs_status)

                if dhw_status := tcs_status.get(SZ_DHW):  # type: ignore[assignment]
                    tcs.hotwater._update_status(dhw_status)  # type: ignore[union-attr]

                for zon_status in tcs_status[SZ_ZONES]:
                    tcs.zones_by_id[zon_status[SZ_ZONE_ID]]._update_status(zon_status)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
"""evohomeasync2 provides an async client for the *updated* Evohome API."""
from __future__ import annotations

from http import HTTPStatus
import logging
from datetime import datetime as dt
from typing import TYPE_CHECKING, NoReturn

import aiohttp

from .broker import Broker
from .controlsystem import ControlSystem
from .exceptions import AuthenticationFailed, DeprecationError, NoSingleTcsError
from .location import Location
from .schema import SCH_FULL_CONFIG, SCH_USER_ACCOUNT


if TYPE_CHECKING:
    from .schema import (
        _FilePathT,
        _LocationIdT,
        _SystemIdT,
        _EvoDictT,
        _EvoListT,
        _EvoSchemaT,
    )


_LOGGER = logging.getLogger(__name__)


class EvohomeClientDeprecated:
    """Deprecated attributes and methods removed from the evohome-client namespace."""

    async def full_installation(
        self, location_id: None | _LocationIdT = None
    ) -> NoReturn:
        # if location_id is None:
        #     location_id = self.installation_info[0]["locationInfo"]["locationId"]
        # url = f"location/{location_id}/installationInfo?"  # Specific location

        raise DeprecationError(
            "EvohomeClient.full_installation() is deprecated, use .installation()"
        )

    async def gateway(self, *args, **kwargs) -> NoReturn:
        raise DeprecationError("EvohomeClient.gateway() is deprecated")

    async def set_status_away(self, *args, **kwargs) -> NoReturn:
        raise DeprecationError(
            "EvohomeClient.set_status_away() is deprecated, use .set_mode_away()"
        )

    async def set_status_custom(self, *args, **kwargs) -> NoReturn:
        raise DeprecationError(
            "EvohomeClient.set_status_custom() is deprecated, use .set_mode_custom()"
        )

    async def set_status_dayoff(self, *args, **kwargs) -> NoReturn:
        raise DeprecationError(
            "EvohomeClient.set_status_dayoff() is deprecated, use .set_mode_dayoff()"
        )

    async def set_status_eco(self, *args, **kwargs) -> NoReturn:
        raise DeprecationError(
            "EvohomeClient.set_status_eco() is deprecated, use .set_mode_eco()"
        )

    async def set_status_heatingoff(self, *args, **kwargs) -> NoReturn:
        raise DeprecationError(
            "EvohomeClient.set_status_heatingoff() is deprecated, use .set_mode_heatingoff()"
        )

    async def set_status_normal(self, *args, **kwargs) -> NoReturn:
        raise DeprecationError(
            "EvohomeClient.set_status_normal() is deprecated, use .set_mode_auto()"
        )

    async def set_status_reset(self, *args, **kwargs) -> NoReturn:
        raise DeprecationError(
            "EvohomeClient.set_status_reset() is deprecated, use .reset_mode()"
        )

    async def zone_schedules_backup(self, *args, **kwargs) -> NoReturn:
        raise DeprecationError(
            "EvohomeClient.zone_schedules_backup() is deprecated, use .backup_schedules()"
        )

    async def zone_schedules_restore(self, *args, **kwargs) -> NoReturn:
        raise DeprecationError(
            "EvohomeClient.zone_schedules_restore() is deprecated, use .restore_schedules()"
        )


class EvohomeClient(EvohomeClientDeprecated):
    """Provide a client to access the Honeywell TCC API."""

    _full_config: _EvoListT = None  # type: ignore[assignment]  # installation_info (all locations of user)
    _user_account: _EvoDictT = None  # type: ignore[assignment]  # account_info

    def __init__(
        self,
        username: str,
        password: str,
        /,
        *,
        debug: bool = False,
        refresh_token: str | None = None,
        access_token: str | None = None,
        access_token_expires: dt | None = None,
        session: None | aiohttp.ClientSession = None,
    ) -> None:
        """Construct the v2 EvohomeClient object.

        If tokens are given then this will be used to try and reduce the number of
        calls to the authentication service which is known to be rate limited.
        """

        self._logger = _LOGGER

        if debug:
            self._logger.setLevel(logging.DEBUG)
            self._logger.debug("Debug mode is explicitly enabled.")

        self._broker = Broker(
            username,
            password,
            refresh_token=refresh_token,
            access_token=access_token,
            access_token_expires=access_token_expires,
            session=session,
            logger=self._logger,
        )

        self.locations: list[Location] = []

    @property
    def username(self) -> str:  # TODO: deprecate? or use config JSON?
        return self._broker._credentials["Username"]

    @property
    def password(self) -> str:  # TODO: deprecate
        return self._broker._credentials["Password"]

    @property
    def refresh_token(self) -> str | None:  # TODO: deprecate
        return self._broker.refresh_token

    @property
    def access_token(self) -> str | None:  # TODO: deprecate
        return self._broker.access_token

    @property
    def access_token_expires(self) -> dt | None:  # TODO: deprecate
        return self._broker.access_token_expires

    async def login(self) -> None:
        """Retrieve the user account and installation details.

        Will authenticate as required.
        """

        try:  # the cached access_token may be valid, but is not authorized
            await self.user_account()

        except AuthenticationFailed as exc:
            if exc.status != HTTPStatus.UNAUTHORIZED or not self.access_token:
                raise

            self._logger.warning(
                "Unauthorized access_token (will try re-authenticating)."
            )
            self._broker.access_token = None  # FIXME: this is a hack
            await self.user_account(force_update=True)

        await self.installation()

    @property  # user_account
    def account_info(self) -> _EvoSchemaT:  # from original evohomeclient namespace
        """Return the information of the user account."""
        return self._user_account

    async def user_account(self, force_update: bool = False) -> _EvoDictT:
        """Return the user account information.

        If required/forced, retrieve that data from the vendor's API.
        """

        # There is usually no need to refresh this data (it is config, not state)
        if self._user_account and not force_update:
            return self._user_account

        self._user_account = await self._broker.get(
            "userAccount", schema=SCH_USER_ACCOUNT
        )  # type: ignore[assignment]

        return self._user_account

    @property  # full_config (all locations of user)
    def installation_info(self) -> _EvoListT:  # from original evohomeclient namespace
        """Return the installation info (config) of all the user's locations."""
        return self._full_config

    async def installation(self, force_update: bool = False) -> _EvoListT:
        """Return the configuration of the user's locations their status.

        If required/forced, retrieve that data from the vendor's API.
        Note that the force_update flag will create new location entities (it includes
        `self.locations = []`).
        """

        # There is usually no need to refresh this data (it is config, not state)
        if self._full_config and not force_update:
            return self._full_config

        return await self._installation()  # aka self.installation_info

    async def _installation(self, refresh_status: bool = True) -> _EvoListT:
        """Return the configuration of the user's locations their status.

        The refresh_status flag is used for dev/test to disable retreiving the initial
        status of each location (and its child entities).
        """

        assert isinstance(self.account_info, dict)  # mypy

        # FIXME: shouldn't really be starting again with new objects?
        self.locations = []  # for now, need to clear this before GET

        url = f"location/installationInfo?userId={self.account_info['userId']}"
        url += "&includeTemperatureControlSystems=True"

        self._full_config = await self._broker.get(url, schema=SCH_FULL_CONFIG)  # type: ignore[assignment]

        # populate each freshly instantiated location with its initial status
        loc_config: _EvoDictT

        for loc_config in self._full_config:
            loc = Location(self, loc_config)
            self.locations.append(loc)
            if refresh_status:
                await loc.refresh_status()

        return self._full_config

    def _get_single_tcs(self) -> ControlSystem:
        """If there is a single location/gateway/TCS, return it, or raise an exception.

        Most users will have only one TCS.
        """

        if not self.locations or len(self.locations) != 1:
            raise NoSingleTcsError(
                "There is not a single location (only) for this account"
            )

        if len(self.locations[0]._gateways) != 1:  # type: ignore[index]
            raise NoSingleTcsError(
                "There is not a single gateway (only) for this account/location"
            )

        if len(self.locations[0]._gateways[0]._control_systems) != 1:  # type: ignore[index]
            raise NoSingleTcsError(
                "There is not a single TCS (only) for this account/location/gateway"
            )

        return self.locations[0]._gateways[0]._control_systems[0]  # type: ignore[index]

    @property
    def system_id(self) -> _SystemIdT:  # an evohome-client anachronism, deprecate?
        """Return the ID of the 'default' TCS (assumes only one loc/gwy/TCS)."""
        return self._get_single_tcs().systemId

    async def reset_mode(self) -> None:
        """Reset the mode of the default TCS and its zones."""
        await self._get_single_tcs().reset_mode()

    async def set_mode_auto(self) -> None:
        """Set the default TCS into auto mode."""
        await self._get_single_tcs().set_auto()

    async def set_mode_away(self, /, *, until: dt | None = None) -> None:
        """Set the default TCS into away mode."""
        await self._get_single_tcs().set_away(until=until)

    async def set_mode_custom(self, /, *, until: dt | None = None) -> None:
        """Set the default TCS into custom mode."""
        await self._get_single_tcs().set_custom(until=until)

    async def set_mode_dayoff(self, /, *, until: dt | None = None) -> None:
        """Set the default TCS into day off mode."""
        await self._get_single_tcs().set_dayoff(until=until)

    async def set_mode_eco(self, /, *, until: dt | None = None) -> None:
        """Set the default TCS into eco mode."""
        await self._get_single_tcs().set_eco(until=until)

    async def set_mode_heatingoff(self, /, *, until: dt | None = None) -> None:
        """Set the default TCS into heating off mode."""
        await self._get_single_tcs().set_heatingoff(until=until)

    async def temperatures(self) -> _EvoListT:
        """Return the current temperatures and setpoints of the default TCS."""
        return await self._get_single_tcs().temperatures()

    async def backup_schedules(self, filename: _FilePathT) -> None:
        """Backup all schedules from the default control system to the file."""
        await self._get_single_tcs().backup_schedules(filename)

    async def restore_schedules(
        self, filename: _FilePathT, match_by_name: bool = False
    ) -> None:
        """Restore all schedules from the file to the control system.

        There is the option to match schedules to their zone/dhw by name rather than id.
        """

        await self._get_single_tcs().restore_schedules(
            filename, match_by_name=match_by_name
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
"""evohomeasync2 - Schema for RESTful API Account JSON."""
from __future__ import annotations

from .const import (
    SZ_CITY,
    SZ_COUNTRY,
    SZ_FIRSTNAME,
    SZ_LANGUAGE,
    SZ_LASTNAME,
    SZ_POSTCODE,
    SZ_STREET_ADDRESS,
    SZ_USER_ID,
    SZ_USERNAME,
)
from .const import obfuscate as _obfuscate
from .helpers import vol  # voluptuous


# These are vendor-specific constants, used for authentication
SZ_ACCESS_TOKEN = "access_token"
SZ_EXPIRES_IN = "expires_in"
SZ_REFRESH_TOKEN = "refresh_token"
SZ_SCOPE = "scope"
SZ_TOKEN_TYPE = "token_type"


SCH_OAUTH_TOKEN = vol.Schema(
    {
        vol.Required(SZ_ACCESS_TOKEN): vol.All(str, _obfuscate),
        vol.Required(SZ_EXPIRES_IN): int,  # 1800 seconds
        vol.Required(SZ_REFRESH_TOKEN): vol.All(str, _obfuscate),
        vol.Required(SZ_TOKEN_TYPE): str,
        vol.Optional(SZ_SCOPE): str,  # "EMEA-V1-Basic EMEA-V1-Anonymous"
    }
)

SCH_USER_ACCOUNT = vol.Schema(
    {
        vol.Required(SZ_USER_ID): str,
        vol.Required(SZ_USERNAME): vol.All(vol.Email(), _obfuscate),
        vol.Required(SZ_FIRSTNAME): str,
        vol.Required(SZ_LASTNAME): vol.All(str, _obfuscate),
        vol.Required(SZ_STREET_ADDRESS): vol.All(str, _obfuscate),
        vol.Required(SZ_CITY): vol.All(str, _obfuscate),
        vol.Required(SZ_POSTCODE): vol.All(str, _obfuscate),
        vol.Required(SZ_COUNTRY): str,
        vol.Required(SZ_LANGUAGE): str,
    },
    extra=vol.PREVENT_EXTRA,
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
"""evohomeasync2 - Constants."""

from enum import EnumCheck, StrEnum, verify
import re
from typing import Final


# all debug flags should be False for published code
_DEBUG_DONT_OBSFUCATE = False  # used for pytest scripts


REGEX_DHW_ID = r"[0-9]*"
REGEX_EMAIL_ADDRESS = r"^([a-zA-Z0-9_\-\.]+)@([a-zA-Z0-9_\-\.]+)\.([a-zA-Z]{2,5})$"
REGEX_GATEWAY_ID = r"[0-9]*"
REGEX_LOCATION_ID = r"[0-9]*"
REGEX_SYSTEM_ID = r"[0-9]*"
REGEX_ZONE_ID = r"[0-9]*"


# These are vendor-specific constants, used for keys
SZ_ACTIVE_FAULTS = "activeFaults"
SZ_ALLOWED_FAN_MODES = "allowedFanModes"
SZ_ALLOWED_MODES = "allowedModes"
SZ_ALLOWED_SETPOINT_MODES = "allowedSetpointModes"
SZ_ALLOWED_STATES = "allowedStates"
SZ_ALLOWED_SYSTEM_MODES = "allowedSystemModes"

SZ_CAN_BE_PERMANENT = "canBePermanent"
SZ_CAN_BE_TEMPORARY = "canBeTemporary"
SZ_CAN_CONTROL_COOL = "canControlCool"
SZ_CAN_CONTROL_HEAT = "canControlHeat"
SZ_CITY = "city"
SZ_COUNTRY = "country"
SZ_CRC = "crc"
SZ_CURRENT_OFFSET_MINUTES = "currentOffsetMinutes"

SZ_DAILY_SCHEDULES = "dailySchedules"
SZ_DAY_OF_WEEK = "dayOfWeek"
SZ_DHW = "dhw"
SZ_DHW_ID = "dhwId"
SZ_DHW_STATE = "dhwState"
SZ_DHW_STATE_CAPABILITIES_RESPONSE = "dhwStateCapabilitiesResponse"
SZ_DISPLAY_NAME = "displayName"
SZ_DOMESTIC_HOT_WATER = "domesticHotWater"

SZ_FAULT_TYPE = "faultType"
SZ_FIRSTNAME = "firstname"

SZ_GATEWAY = "gateway"
SZ_GATEWAY_ID = "gatewayId"
SZ_GATEWAY_INFO = "gatewayInfo"
SZ_GATEWAYS = "gateways"

SZ_HEAT_SETPOINT = "heatSetpoint"
SZ_HEAT_SETPOINT_VALUE = "HeatSetpointValue"
SZ_HEATING_ZONE = "HeatingZone"

SZ_IS_AVAILABLE = "isAvailable"
SZ_IS_CANCELABLE = "isCancelable"
SZ_IS_CHANGEABLE = "isChangeable"
SZ_IS_PERMANENT = "isPermanent"
SZ_IS_WI_FI = "isWiFi"

SZ_LANGUAGE = "language"
SZ_LASTNAME = "lastname"
SZ_LOCATION = "location"
SZ_LOCATION_ID = "locationId"
SZ_LOCATION_INFO = "locationInfo"
SZ_LOCATION_OWNER = "locationOwner"
SZ_LOCATION_TYPE = "locationType"

SZ_MAC = "mac"
SZ_MAX_COOL_SETPOINT = "maxCoolSetpoint"
SZ_MAX_DURATION = "maxDuration"
SZ_MAX_HEAT_SETPOINT = "maxHeatSetpoint"
SZ_MAX_SWITCHPOINTS_PER_DAY = "maxSwitchpointsPerDay"
SZ_MIN_COOL_SETPOINT = "minCoolSetpoint"
SZ_MIN_HEAT_SETPOINT = "minHeatSetpoint"
SZ_MIN_SWITCHPOINTS_PER_DAY = "minSwitchpointsPerDay"
SZ_MODE = "mode"
SZ_MODEL_TYPE = "modelType"

SZ_NAME = "name"

SZ_OFFSET_MINUTES = "offsetMinutes"

SZ_PERMANENT = "permanent"
SZ_POSTCODE = "postcode"

SZ_SCHEDULE_CAPABILITIES = "scheduleCapabilities"
SZ_SCHEDULE_CAPABILITIES_RESPONSE = "scheduleCapabilitiesResponse"
SZ_SETPOINT_CAPABILITIES = "setpointCapabilities"
SZ_SETPOINT_MODE = "setpointMode"
SZ_SETPOINT_STATUS = "setpointStatus"
SZ_SETPOINT_VALUE_RESOLUTION = "setpointValueResolution"
SZ_SINCE = "since"
SZ_STATE = "state"
SZ_STATE_STATUS = "stateStatus"
SZ_STREET_ADDRESS = "streetAddress"
SZ_SUPPORTS_DAYLIGHT_SAVING = "supportsDaylightSaving"
SZ_SWITCHPOINTS = "switchpoints"
SZ_SYSTEM_ID = "systemId"
SZ_SYSTEM_MODE = "systemMode"
SZ_SYSTEM_MODE_STATUS = "systemModeStatus"

SZ_TARGET_HEAT_TEMPERATURE = "targetHeatTemperature"
SZ_TEMPERATURE = "temperature"
SZ_TEMPERATURE_CONTROL_SYSTEM = "temperatureControlSystem"
SZ_TEMPERATURE_CONTROL_SYSTEMS = "temperatureControlSystems"
SZ_TEMPERATURE_STATUS = "temperatureStatus"
SZ_TEMPERATURE_ZONE = "temperatureZone"
SZ_TIME_OF_DAY = "timeOfDay"
SZ_TIME_UNTIL = "timeUntil"
SZ_TIME_ZONE = "timeZone"
SZ_TIME_ZONE_ID = "timeZoneId"
SZ_TIMING_MODE = "timingMode"
SZ_TIMING_RESOLUTION = "timingResolution"

SZ_UNTIL = "until"
SZ_UNTIL_TIME = "untilTime"
SZ_USE_DAYLIGHT_SAVE_SWITCHING = "useDaylightSaveSwitching"
SZ_USER_ACCOUNT = "userAccount"
SZ_USER_ID = "userId"
SZ_USERNAME = "username"

SZ_VACATION_HOLD_CAPABILITIES = "vacationHoldCapabilities"
SZ_VALUE_RESOLUTION = "valueResolution"

SZ_ZONE_ID = "zoneId"
SZ_ZONE_TYPE = "zoneType"
SZ_ZONES = "zones"


SZ_MONDAY = "Monday"
SZ_TUESDAY = "Tuesday"
SZ_WEDNESDAY = "Wednesday"
SZ_THURSDAY = "Thursday"
SZ_FRIDAY = "Friday"
SZ_SATURDAY = "Saturday"
SZ_SUNDAY = "Sunday"

DAYS_OF_WEEK = (
    SZ_MONDAY,
    SZ_TUESDAY,
    SZ_WEDNESDAY,
    SZ_THURSDAY,
    SZ_FRIDAY,
    SZ_SATURDAY,
    SZ_SUNDAY,
)


SZ_OFF = "Off"
SZ_ON = "On"


@verify(EnumCheck.UNIQUE)
class DhwState(StrEnum):
    OFF: Final[str] = SZ_OFF
    ON: Final[str] = SZ_ON


@verify(EnumCheck.UNIQUE)
class FaultType(StrEnum):
    GCL__: Final[str] = "GatewayCommunicationLost"
    TZACL: Final[str] = "TempZoneActuatorCommunicationLost"
    TZALB: Final[str] = "TempZoneActuatorLowBattery"
    TZSCL: Final[str] = "TempZoneSensorCommunicationLost"
    TZSLB: Final[str] = "TempZoneSensorLowBattery"


SZ_AUTO = "Auto"
SZ_AUTO_WITH_ECO = "AutoWithEco"
SZ_AUTO_WITH_RESET = "AutoWithReset"
SZ_AWAY = "Away"
SZ_CUSTOM = "Custom"
SZ_DAY_OFF = "DayOff"
SZ_HEATING_OFF = "HeatingOff"

SZ_COOL = "Cool"
SZ_HEAT = "Heat"


@verify(EnumCheck.UNIQUE)
class SystemMode(StrEnum):
    AUTO: Final[str] = SZ_AUTO
    AUTO_WITH_ECO: Final[str] = SZ_AUTO_WITH_ECO
    AUTO_WITH_RESET: Final[str] = SZ_AUTO_WITH_RESET
    AWAY: Final[str] = SZ_AWAY
    CUSTOM: Final[str] = SZ_CUSTOM
    DAY_OFF: Final[str] = SZ_DAY_OFF
    HEATING_OFF: Final[str] = SZ_HEATING_OFF
    OFF: Final[str] = SZ_OFF  # not evohome (VisionProWifiRetail)
    HEAT: Final[str] = SZ_HEAT  # not evohome (VisionProWifiRetail)


@verify(EnumCheck.UNIQUE)
class TcsModelType(StrEnum):
    EVO_TOUCH: Final[str] = "EvoTouch"
    FOCUS_PRO_WIFI_RETAIL: Final[str] = "FocusProWifiRetail"
    VISION_PRO_WIFI_RETAIL: Final[str] = "VisionProWifiRetail"


SZ_FOLLOW_SCHEDULE = "FollowSchedule"
SZ_PERMANENT_OVERRIDE = "PermanentOverride"
SZ_TEMPORARY_OVERRIDE = "TemporaryOverride"
SZ_VACATION_HOLD = "VacationHold"


@verify(EnumCheck.UNIQUE)
class ZoneMode(StrEnum):
    FOLLOW_SCHEDULE: Final[str] = SZ_FOLLOW_SCHEDULE
    PERMANENT_OVERRIDE: Final[str] = SZ_PERMANENT_OVERRIDE
    TEMPORARY_OVERRIDE: Final[str] = SZ_TEMPORARY_OVERRIDE
    VACATION_HOLD: Final[str] = SZ_VACATION_HOLD  # not evohome (VisionProWifiRetail)


@verify(EnumCheck.UNIQUE)
class ZoneModelType(StrEnum):
    FOCUS_PRO_WIFI_RETAIL: Final[str] = "FocusProWifiRetail"
    HEATING_ZONE: Final[str] = "HeatingZone"
    ROUND_MODULATION: Final[str] = "RoundModulation"
    ROUND_WIRELESS: Final[str] = "RoundWireless"
    UNKNOWN: Final[str] = "Unknown"
    VISION_PRO_WIFI_RETAIL: Final[str] = "VisionProWifiRetail"


SZ_ELECTRIC_HEAT = "ElectricHeat"  # TODO: needs confirming
SZ_MIXING_VALVE = "MixingValve"
SZ_RADIATOR_ZONE = "RadiatorZone"
SZ_THERMOSTAT = "Thermostat"
SZ_UNDERFLOOR_HEATING = "UnderfloorHeating"
SZ_UNKNOWN = "Unknown"
SZ_ZONE_VALVES = "ZoneValves"  # is not ZoneValve


@verify(EnumCheck.UNIQUE)
class ZoneType(StrEnum):
    MIXING_VALVE: Final[str] = SZ_MIXING_VALVE
    RADIATOR_ZONE: Final[str] = SZ_RADIATOR_ZONE
    THERMOSTAT: Final[str] = SZ_THERMOSTAT
    UNDERFLOOR_HEATING: Final[str] = SZ_UNDERFLOOR_HEATING
    UNKNOWN: Final[str] = SZ_UNKNOWN
    ZONE_VALVES: Final[str] = SZ_ZONE_VALVES


# these may not be required with Python 3.12+ (used for 'if mode in ZONE_MODES'...)
DHW_STATES = tuple(x.value for x in DhwState)
FAULT_TYPES = tuple(x.value for x in FaultType)
SYSTEM_MODES = tuple(x.value for x in SystemMode)
TCS_MODEL_TYPES = tuple(x.value for x in TcsModelType)
ZONE_MODEL_TYPES = tuple(x.value for x in ZoneModelType)
ZONE_MODES = tuple(x.value for x in ZoneMode)
ZONE_TYPES = tuple(x.value for x in ZoneType)


def obfuscate(value: bool | int | str):
    if _DEBUG_DONT_OBSFUCATE:
        return value
    if isinstance(value, bool):
        return None
    if isinstance(value, int):
        return 0
    if not isinstance(value, str):
        raise TypeError(f"obfuscate() expects bool | int | str, got {type(value)}")
    if re.match(REGEX_EMAIL_ADDRESS, value):
        return "nobody@nowhere.com"
    return "********"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
"""evohomeasync2 provides an async client for the updated Evohome API."""
from __future__ import annotations


class EvohomeBaseError(Exception):
    """The base exception class for evohome-async."""

    def __init__(self, message: str) -> None:
        super().__init__(message)
        self.message = message


class EvohomeError(EvohomeBaseError):
    """The base exception class for evohome-async."""


class DeprecationError(EvohomeBaseError):
    """The method or property has changed, or is otherwise deprecated."""


class InvalidSchema(EvohomeError):
    """The config/status JSON is invalid (e.g. missing an entity Id)."""


class InvalidParameter(EvohomeError):
    """The supplied parameter(s) is/are invalid (e.g. unknown/unsupported mode)."""


class InvalidSchedule(InvalidSchema):
    """The schedule has an invalid schema."""


class NoSingleTcsError(EvohomeError):
    """There is not exactly one TCS in the user's installation."""


class RequestFailed(EvohomeError):
    """The API request failed for some reason (no/invalid/unexpected response).

    Could be caused by any aiohttp.ClientError, for example: ConnectionError.  If the
    cause was a ClientResponseError, then the `status` attr will have an integer value.
    """

    def __init__(self, message: str, status: None | int = None) -> None:
        super().__init__(message)
        self.status = status  # iff cause was aiohttp.ClientResponseError


class RateLimitExceeded(RequestFailed):
    """API request failed because the vendor's API rate limit was exceeded."""


class AuthenticationFailed(RequestFailed):
    """Unable to authenticate (unable to obtain an access token).

    The cause could be any FailedRequest, including RateLimitExceeded.
    """



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
"""evohomeasync provides an async client for the *original* Evohome API."""

from typing import Any

# TCC config, status dicts
_EvoLeafT = bool | float | int | str | list[str]  # Any
_EvoDictT = dict[str, Any]  # '_EvoDictT' | _EvoLeafT]
_EvoListT = list[_EvoDictT]
_EvoSchemaT = _EvoDictT | _EvoListT

# TCC identifiers
_DhwIdT = str
_GatewayIdT = str
_LocationIdT = str
_SystemIdT = str
_ZoneIdT = str

_DeviceIdT = _DhwIdT | _ZoneIdT

# TCC other
_ModeT = str
_SystemModeT = str

_TaskIdT = str



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
"""evohomeasync2 provides an async client for the updated Evohome API."""

from .schema import (  # noqa: F401
    DhwState,
    SystemMode,
    ZoneMode,
)

URL_HOST = "https://tccna.honeywell.com"

AUTH_URL = f"{URL_HOST}/Auth/OAuth/Token"
URL_BASE = f"{URL_HOST}/WebAPI/emea/api/v1"

HDR_STRFTIME = "%Y-%m-%d %H:%M:%S"  # used by HTTP headers
API_STRFTIME = "%Y-%m-%dT%H:%M:%SZ"  # used by API

AUTH_HEADER_ACCEPT = (
    "application/json, application/xml, "
    "text/json, text/x-json, "
    "text/javascript, text/xml"
)

_AUTH_HEADER_BASIC = (
    "Basic "
    "NGEyMzEwODktZDJiNi00MWJkLWE1ZWItMTZhMGE0MjJiOTk5OjFhMTVjZGI4LTQyZGUtNDA3Y"
    "i1hZGQwLTA1OWY5MmM1MzBjYg=="
)

AUTH_HEADER = {
    "Accept": AUTH_HEADER_ACCEPT,
    "Authorization": _AUTH_HEADER_BASIC,
}

AUTH_PAYLOAD = {
    "Content-Type": "application/x-www-form-urlencoded; charset=utf-8",
    "Host": "rs.alarmnet.com/",
    "Cache-Control": "no-store no-cache",
    "Pragma": "no-cache",
    "Connection": "Keep-Alive",
}

CREDS_REFRESH_TOKEN = {
    "grant_type": "refresh_token",
    "scope": "EMEA-V1-Basic EMEA-V1-Anonymous",
    "refresh_token": None,
}

CREDS_USER_PASSWORD = {
    "grant_type": "password",
    "scope": "EMEA-V1-Basic EMEA-V1-Anonymous EMEA-V1-Get-Current-User-Account",
    "Username": None,
    "Password": None,
}



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
"""evohome-async - validate the evohome-async APIs (methods)."""
from __future__ import annotations

import aiohttp
from datetime import datetime as dt
import pytest
import pytest_asyncio

import evohomeasync2 as evo

from . import mocked_server as mock
from evohomeasync2.schema import (
    SCH_DHW_STATUS,
    SCH_FULL_CONFIG,
    SCH_LOCN_STATUS,
    SCH_USER_ACCOUNT,
    SCH_ZONE_STATUS,
)
from evohomeasync2.schema import SystemMode, SYSTEM_MODES
from evohomeasync2.schema.const import SZ_MODE
from evohomeasync2.schema.schedule import SCH_PUT_SCHEDULE_DHW, SCH_PUT_SCHEDULE_ZONE

from . import _DEBUG_USE_REAL_AIOHTTP
from .helpers import user_credentials as _user_credentials
from .helpers import client_session as _client_session
from .helpers import extract_oauth_tokens


_global_oauth_tokens: tuple[str, str, dt] = None, None, None


@pytest.fixture()
def user_credentials():
    return _user_credentials()


@pytest_asyncio.fixture
async def session():
    client_session = _client_session()
    try:
        yield client_session
    finally:
        await client_session.close()


async def _test_basics_apis(
    username: str,
    password: str,
    session: aiohttp.ClientSession | mock.ClientSession | None = None,
):
    """Test authentication, `user_account()` and `installation()`."""

    global _global_oauth_tokens

    refresh_token, access_token, access_token_expires = _global_oauth_tokens

    #
    # STEP 0: Instantiation, NOTE: No API calls invoked during instantiation
    client = evo.EvohomeClient(
        username,
        password,
        session=session,
        refresh_token=refresh_token,
        access_token=access_token,
        access_token_expires=access_token_expires,
    )

    #
    # STEP 1: Authentication (isolated from client.login()), POST /Auth/OAuth/Token
    await client._broker._basic_login()
    _global_oauth_tokens = extract_oauth_tokens(client)

    assert isinstance(client.access_token, str)
    assert isinstance(client.access_token_expires, dt)
    assert isinstance(client.refresh_token, str)

    access_token = client.access_token
    refresh_token = client.refresh_token

    await client._broker._headers()

    # The above should not cause a re-authentication, so...
    assert client.access_token == access_token
    assert client.refresh_token == refresh_token

    #
    # STEP 1A: Re-Authentication: more likely to cause 429: Too Many Requests
    if isinstance(session, aiohttp.ClientSession):
        access_token = client.access_token
        refresh_token = client.refresh_token

    #     await client._basic_login()  # re-authenticate using refresh_token

    #     assert True or client.access_token != access_token  # TODO: mocked_server wont do this
    #     assert True or client.refresh_token != refresh_token  # TODO: mocked_server wont do this

    #
    # STEP 2: User account,  GET /userAccount...
    assert client.account_info is None

    await client.user_account(force_update=False)  # will update as no access_token

    assert SCH_USER_ACCOUNT(client._user_account)
    assert client.account_info == client._user_account

    await client.user_account()  # wont update as access_token is valid
    # await client.user_account(force_update=True)  # will update as forced

    #
    # STEP 3: Installation, GET /location/installationInfo?userId={userId}
    assert client.locations == []
    assert client.installation_info is None

    await client._installation(refresh_status=False)  # not client.installation()

    assert SCH_FULL_CONFIG(client._full_config)  # an array of locations
    assert client.installation_info == client._full_config

    # assert isinstance(client.system_id, str)  # only if one TCS
    assert client.locations

    await client.installation()  # not client._installation()
    # await client.installation(force_update=True)  # will update as forced

    #
    # STEP 4: Status, GET /location/{locationId}/status
    for loc in client.locations:
        loc_status = await loc.refresh_status()
        assert SCH_LOCN_STATUS(loc_status)

    pass


async def _test_sched__apis(
    username: str,
    password: str,
    session: aiohttp.ClientSession | mock.ClientSession | None = None,
):
    """Test `get_schedule()` and `get_schedule()`."""

    global _global_oauth_tokens

    refresh_token, access_token, access_token_expires = _global_oauth_tokens

    #
    # STEP 0: Instantiation...
    client = evo.EvohomeClient(
        username,
        password,
        session=session,
        refresh_token=refresh_token,
        access_token=access_token,
        access_token_expires=access_token_expires,
    )

    #
    # STEP 1: Initial authentication, retrieve config & status
    await client.login()  # invokes await client.installation()
    _global_oauth_tokens = extract_oauth_tokens(client)

    #
    # STEP 2: GET & PUT /{_type}/{_id}/schedule
    if dhw := client._get_single_tcs().hotwater:
        schedule = await dhw.get_schedule()
        assert SCH_PUT_SCHEDULE_DHW(schedule)
        await dhw.set_schedule(schedule)

    if zone := client._get_single_tcs()._zones[0]:
        schedule = await zone.get_schedule()
        assert SCH_PUT_SCHEDULE_ZONE(schedule)
        await zone.set_schedule(schedule)

    pass


async def _test_status_apis(
    username: str,
    password: str,
    session: aiohttp.ClientSession | mock.ClientSession | None = None,
):
    """Test `_refresh_status()` for DHW/zone."""

    global _global_oauth_tokens

    refresh_token, access_token, access_token_expires = _global_oauth_tokens

    #
    # STEP 0: Instantiation...
    client = evo.EvohomeClient(
        username,
        password,
        session=session,
        refresh_token=refresh_token,
        access_token=access_token,
        access_token_expires=access_token_expires,
    )

    #
    # STEP 1: Initial authentication, retrieve config & status
    await client.login()  # invokes await client.installation()
    _global_oauth_tokens = extract_oauth_tokens(client)

    #
    # STEP 2: GET /{_type}/{_id}/status
    if dhw := client._get_single_tcs().hotwater:
        dhw_status = await dhw._refresh_status()
        assert SCH_DHW_STATUS(dhw_status)

    if zone := client._get_single_tcs()._zones[0]:
        zone_status = await zone._refresh_status()
        assert SCH_ZONE_STATUS(zone_status)

    pass


async def _test_system_apis(
    username: str,
    password: str,
    session: aiohttp.ClientSession | mock.ClientSession | None = None,
):
    """Test `set_mode()` for TCS."""
    global _global_oauth_tokens

    refresh_token, access_token, access_token_expires = _global_oauth_tokens

    #
    # STEP 0: Instantiation...
    client = evo.EvohomeClient(
        username,
        password,
        session=session,
        refresh_token=refresh_token,
        access_token=access_token,
        access_token_expires=access_token_expires,
    )

    #
    # STEP 1: Initial authentication, retrieve config & status
    await client.login()  # invokes await client.installation()
    _global_oauth_tokens = extract_oauth_tokens(client)

    #
    # STEP 2: GET /{_type}/{_id}/status
    try:
        tcs = client._get_single_tcs()
    except evo.NoSingleTcsError:
        tcs = client.locations[0].gateways[0].control_systems[0]

    mode = tcs.systemModeStatus[SZ_MODE]
    assert mode in SYSTEM_MODES

    await tcs.set_mode(SystemMode.AWAY)
    await client._installation(refresh_status=True)

    await tcs.set_mode(mode)

    pass


@pytest.mark.asyncio
async def test_basics(
    user_credentials: tuple[str, str],
    session: aiohttp.ClientSession | mock.ClientSession,
):
    username, password = user_credentials

    try:
        await _test_basics_apis(username, password, session=session)
    except evo.AuthenticationFailed:
        if not _DEBUG_USE_REAL_AIOHTTP:
            raise
        pytest.skip("Unable to authenticate")


@pytest.mark.asyncio
async def test_sched_(
    user_credentials: tuple[str, str],
    session: aiohttp.ClientSession | mock.ClientSession,
):
    username, password = user_credentials

    try:
        await _test_sched__apis(username, password, session=session)
    except evo.AuthenticationFailed:
        if not _DEBUG_USE_REAL_AIOHTTP:
            raise
        pytest.skip("Unable to authenticate")


@pytest.mark.asyncio
async def test_status(
    user_credentials: tuple[str, str],
    session: aiohttp.ClientSession | mock.ClientSession,
):
    username, password = user_credentials

    try:
        await _test_status_apis(username, password, session=session)
    except evo.AuthenticationFailed:
        if not _DEBUG_USE_REAL_AIOHTTP:
            raise
        pytest.skip("Unable to authenticate")


@pytest.mark.asyncio
async def test_system(
    user_credentials: tuple[str, str],
    session: aiohttp.ClientSession | mock.ClientSession,
):
    username, password = user_credentials

    try:
        await _test_system_apis(username, password, session=session)
    except NotImplementedError:  # TODO: implement
        if _DEBUG_USE_REAL_AIOHTTP:
            raise
        pytest.skip("Mocked server API not implemented")
    except evo.AuthenticationFailed:
        if not _DEBUG_USE_REAL_AIOHTTP:
            raise
        pytest.skip("Unable to authenticate")



schedule = {}
schedule["Kitchen"] = {
    "DailySchedules": [
        {
            "DayOfWeek": 0,
            "Switchpoints": [
                {"heatSetpoint": 19.0, "TimeOfDay": "06:30:00"},
                {"heatSetpoint": 16.0, "TimeOfDay": "09:00:00"},
                {"heatSetpoint": 19.0, "TimeOfDay": "17:30:00"},
                {"heatSetpoint": 16.0, "TimeOfDay": "23:00:00"},
            ],
        },
        {
            "DayOfWeek": 1,
            "Switchpoints": [
                {"heatSetpoint": 19.0, "TimeOfDay": "06:30:00"},
                {"heatSetpoint": 16.0, "TimeOfDay": "09:00:00"},
                {"heatSetpoint": 19.0, "TimeOfDay": "17:30:00"},
                {"heatSetpoint": 16.0, "TimeOfDay": "23:00:00"},
            ],
        },
        {
            "DayOfWeek": 2,
            "Switchpoints": [
                {"heatSetpoint": 19.0, "TimeOfDay": "06:30:00"},
                {"heatSetpoint": 16.0, "TimeOfDay": "09:00:00"},
                {"heatSetpoint": 19.0, "TimeOfDay": "17:30:00"},
                {"heatSetpoint": 16.0, "TimeOfDay": "23:00:00"},
            ],
        },
        {
            "DayOfWeek": 3,
            "Switchpoints": [
                {"heatSetpoint": 19.0, "TimeOfDay": "06:30:00"},
                {"heatSetpoint": 16.0, "TimeOfDay": "09:00:00"},
                {"heatSetpoint": 19.0, "TimeOfDay": "17:30:00"},
                {"heatSetpoint": 16.0, "TimeOfDay": "23:00:00"},
            ],
        },
        {
            "DayOfWeek": 4,
            "Switchpoints": [
                {"heatSetpoint": 19.0, "TimeOfDay": "06:30:00"},
                {"heatSetpoint": 16.0, "TimeOfDay": "09:00:00"},
                {"heatSetpoint": 19.0, "TimeOfDay": "17:30:00"},
                {"heatSetpoint": 16.0, "TimeOfDay": "23:00:00"},
            ],
        },
        {
            "DayOfWeek": 5,
            "Switchpoints": [
                {"heatSetpoint": 19.0, "TimeOfDay": "06:30:00"},
                {"heatSetpoint": 19.0, "TimeOfDay": "17:30:00"},
                {"heatSetpoint": 16.0, "TimeOfDay": "23:00:00"},
            ],
        },
        {
            "DayOfWeek": 6,
            "Switchpoints": [
                {"heatSetpoint": 19.0, "TimeOfDay": "06:30:00"},
                {"heatSetpoint": 19.0, "TimeOfDay": "17:30:00"},
                {"heatSetpoint": 16.0, "TimeOfDay": "23:00:00"},
            ],
        },
    ]
}
schedule["Basement One"] = {
    "DailySchedules": [
        {
            "DayOfWeek": 0,
            "Switchpoints": [{"heatSetpoint": 16.0, "TimeOfDay": "06:00:00"}],
        },
        {
            "DayOfWeek": 1,
            "Switchpoints": [{"heatSetpoint": 16.0, "TimeOfDay": "06:00:00"}],
        },
        {
            "DayOfWeek": 2,
            "Switchpoints": [{"heatSetpoint": 16.0, "TimeOfDay": "06:00:00"}],
        },
        {
            "DayOfWeek": 3,
            "Switchpoints": [{"heatSetpoint": 16.0, "TimeOfDay": "06:00:00"}],
        },
        {
            "DayOfWeek": 4,
            "Switchpoints": [{"heatSetpoint": 16.0, "TimeOfDay": "06:00:00"}],
        },
        {
            "DayOfWeek": 5,
            "Switchpoints": [{"heatSetpoint": 16.0, "TimeOfDay": "06:00:00"}],
        },
        {
            "DayOfWeek": 6,
            "Switchpoints": [{"heatSetpoint": 16.0, "TimeOfDay": "06:00:00"}],
        },
    ]
}
schedule["Dining room"] = {
    "DailySchedules": [
        {
            "DayOfWeek": 0,
            "Switchpoints": [
                {"heatSetpoint": 19.0, "TimeOfDay": "06:30:00"},
                {"heatSetpoint": 14.0, "TimeOfDay": "09:00:00"},
                {"heatSetpoint": 21.0, "TimeOfDay": "17:30:00"},
                {"heatSetpoint": 14.0, "TimeOfDay": "23:00:00"},
            ],
        },
        {
            "DayOfWeek": 1,
            "Switchpoints": [
                {"heatSetpoint": 19.0, "TimeOfDay": "06:30:00"},
                {"heatSetpoint": 14.0, "TimeOfDay": "09:00:00"},
                {"heatSetpoint": 21.0, "TimeOfDay": "17:30:00"},
                {"heatSetpoint": 14.0, "TimeOfDay": "23:00:00"},
            ],
        },
        {
            "DayOfWeek": 2,
            "Switchpoints": [
                {"heatSetpoint": 19.0, "TimeOfDay": "06:30:00"},
                {"heatSetpoint": 14.0, "TimeOfDay": "09:00:00"},
                {"heatSetpoint": 21.0, "TimeOfDay": "17:30:00"},
                {"heatSetpoint": 14.0, "TimeOfDay": "23:00:00"},
            ],
        },
        {
            "DayOfWeek": 3,
            "Switchpoints": [
                {"heatSetpoint": 19.0, "TimeOfDay": "06:30:00"},
                {"heatSetpoint": 14.0, "TimeOfDay": "09:00:00"},
                {"heatSetpoint": 21.0, "TimeOfDay": "17:30:00"},
                {"heatSetpoint": 14.0, "TimeOfDay": "23:00:00"},
            ],
        },
        {
            "DayOfWeek": 4,
            "Switchpoints": [
                {"heatSetpoint": 19.0, "TimeOfDay": "06:30:00"},
                {"heatSetpoint": 14.0, "TimeOfDay": "09:00:00"},
                {"heatSetpoint": 21.0, "TimeOfDay": "17:30:00"},
                {"heatSetpoint": 14.0, "TimeOfDay": "23:00:00"},
            ],
        },
        {
            "DayOfWeek": 5,
            "Switchpoints": [
                {"heatSetpoint": 19.0, "TimeOfDay": "06:30:00"},
                {"heatSetpoint": 21.0, "TimeOfDay": "17:30:00"},
                {"heatSetpoint": 14.0, "TimeOfDay": "23:00:00"},
            ],
        },
        {
            "DayOfWeek": 6,
            "Switchpoints": [
                {"heatSetpoint": 19.0, "TimeOfDay": "06:30:00"},
                {"heatSetpoint": 21.0, "TimeOfDay": "17:30:00"},
                {"heatSetpoint": 14.0, "TimeOfDay": "23:00:00"},
            ],
        },
    ]
}
schedule["Landing"] = {
    "DailySchedules": [
        {
            "DayOfWeek": 0,
            "Switchpoints": [{"heatSetpoint": 16.0, "TimeOfDay": "06:00:00"}],
        },
        {
            "DayOfWeek": 1,
            "Switchpoints": [{"heatSetpoint": 16.0, "TimeOfDay": "06:00:00"}],
        },
        {
            "DayOfWeek": 2,
            "Switchpoints": [{"heatSetpoint": 16.0, "TimeOfDay": "06:00:00"}],
        },
        {
            "DayOfWeek": 3,
            "Switchpoints": [{"heatSetpoint": 16.0, "TimeOfDay": "06:00:00"}],
        },
        {
            "DayOfWeek": 4,
            "Switchpoints": [{"heatSetpoint": 16.0, "TimeOfDay": "06:00:00"}],
        },
        {
            "DayOfWeek": 5,
            "Switchpoints": [{"heatSetpoint": 16.0, "TimeOfDay": "06:00:00"}],
        },
        {
            "DayOfWeek": 6,
            "Switchpoints": [{"heatSetpoint": 16.0, "TimeOfDay": "06:00:00"}],
        },
    ]
}
schedule["Guest Bedroom"] = {
    "DailySchedules": [
        {
            "DayOfWeek": 0,
            "Switchpoints": [
                {"heatSetpoint": 14.0, "TimeOfDay": "00:00:00"},
                {"heatSetpoint": 14.0, "TimeOfDay": "06:00:00"},
                {"heatSetpoint": 14.0, "TimeOfDay": "08:30:00"},
                {"heatSetpoint": 14.0, "TimeOfDay": "20:00:00"},
            ],
        },
        {
            "DayOfWeek": 1,
            "Switchpoints": [
                {"heatSetpoint": 14.0, "TimeOfDay": "00:00:00"},
                {"heatSetpoint": 14.0, "TimeOfDay": "06:00:00"},
                {"heatSetpoint": 14.0, "TimeOfDay": "08:30:00"},
                {"heatSetpoint": 14.0, "TimeOfDay": "20:00:00"},
            ],
        },
        {
            "DayOfWeek": 2,
            "Switchpoints": [
                {"heatSetpoint": 14.0, "TimeOfDay": "00:00:00"},
                {"heatSetpoint": 14.0, "TimeOfDay": "06:00:00"},
                {"heatSetpoint": 14.0, "TimeOfDay": "08:30:00"},
                {"heatSetpoint": 14.0, "TimeOfDay": "20:00:00"},
            ],
        },
        {
            "DayOfWeek": 3,
            "Switchpoints": [
                {"heatSetpoint": 14.0, "TimeOfDay": "00:00:00"},
                {"heatSetpoint": 14.0, "TimeOfDay": "06:00:00"},
                {"heatSetpoint": 14.0, "TimeOfDay": "08:30:00"},
                {"heatSetpoint": 14.0, "TimeOfDay": "20:00:00"},
            ],
        },
        {
            "DayOfWeek": 4,
            "Switchpoints": [
                {"heatSetpoint": 14.0, "TimeOfDay": "00:00:00"},
                {"heatSetpoint": 14.0, "TimeOfDay": "06:00:00"},
                {"heatSetpoint": 14.0, "TimeOfDay": "08:30:00"},
                {"heatSetpoint": 14.0, "TimeOfDay": "20:00:00"},
            ],
        },
        {
            "DayOfWeek": 5,
            "Switchpoints": [
                {"heatSetpoint": 14.0, "TimeOfDay": "00:00:00"},
                {"heatSetpoint": 14.0, "TimeOfDay": "06:00:00"},
                {"heatSetpoint": 14.0, "TimeOfDay": "08:30:00"},
                {"heatSetpoint": 14.0, "TimeOfDay": "20:00:00"},
            ],
        },
        {
            "DayOfWeek": 6,
            "Switchpoints": [
                {"heatSetpoint": 14.0, "TimeOfDay": "00:00:00"},
                {"heatSetpoint": 14.0, "TimeOfDay": "06:00:00"},
                {"heatSetpoint": 14.0, "TimeOfDay": "08:30:00"},
                {"heatSetpoint": 14.0, "TimeOfDay": "20:00:00"},
            ],
        },
    ]
}
schedule["Sues Bedroom"] = {
    "DailySchedules": [
        {
            "DayOfWeek": 0,
            "Switchpoints": [
                {"heatSetpoint": 14.0, "TimeOfDay": "00:00:00"},
                {"heatSetpoint": 20.0, "TimeOfDay": "06:00:00"},
                {"heatSetpoint": 14.0, "TimeOfDay": "08:30:00"},
                {"heatSetpoint": 20.0, "TimeOfDay": "20:00:00"},
            ],
        },
        {
            "DayOfWeek": 1,
            "Switchpoints": [
                {"heatSetpoint": 14.0, "TimeOfDay": "00:00:00"},
                {"heatSetpoint": 20.0, "TimeOfDay": "06:00:00"},
                {"heatSetpoint": 14.0, "TimeOfDay": "08:30:00"},
                {"heatSetpoint": 20.0, "TimeOfDay": "20:00:00"},
            ],
        },
        {
            "DayOfWeek": 2,
            "Switchpoints": [
                {"heatSetpoint": 14.0, "TimeOfDay": "00:00:00"},
                {"heatSetpoint": 20.0, "TimeOfDay": "06:00:00"},
                {"heatSetpoint": 14.0, "TimeOfDay": "08:30:00"},
                {"heatSetpoint": 20.0, "TimeOfDay": "20:00:00"},
            ],
        },
        {
            "DayOfWeek": 3,
            "Switchpoints": [
                {"heatSetpoint": 14.0, "TimeOfDay": "00:00:00"},
                {"heatSetpoint": 20.0, "TimeOfDay": "06:00:00"},
                {"heatSetpoint": 14.0, "TimeOfDay": "08:30:00"},
                {"heatSetpoint": 20.0, "TimeOfDay": "20:00:00"},
            ],
        },
        {
            "DayOfWeek": 4,
            "Switchpoints": [
                {"heatSetpoint": 14.0, "TimeOfDay": "00:00:00"},
                {"heatSetpoint": 20.0, "TimeOfDay": "06:00:00"},
                {"heatSetpoint": 14.0, "TimeOfDay": "08:30:00"},
                {"heatSetpoint": 20.0, "TimeOfDay": "20:00:00"},
            ],
        },
        {
            "DayOfWeek": 5,
            "Switchpoints": [
                {"heatSetpoint": 14.0, "TimeOfDay": "00:00:00"},
                {"heatSetpoint": 20.0, "TimeOfDay": "06:00:00"},
                {"heatSetpoint": 14.0, "TimeOfDay": "08:30:00"},
                {"heatSetpoint": 20.0, "TimeOfDay": "20:00:00"},
            ],
        },
        {
            "DayOfWeek": 6,
            "Switchpoints": [
                {"heatSetpoint": 14.0, "TimeOfDay": "00:00:00"},
                {"heatSetpoint": 20.0, "TimeOfDay": "06:00:00"},
                {"heatSetpoint": 14.0, "TimeOfDay": "08:30:00"},
                {"heatSetpoint": 20.0, "TimeOfDay": "20:00:00"},
            ],
        },
    ]
}
schedule["Matts bedroom"] = {
    "DailySchedules": [
        {
            "DayOfWeek": 0,
            "Switchpoints": [
                {"heatSetpoint": 16.0, "TimeOfDay": "00:00:00"},
                {"heatSetpoint": 19.0, "TimeOfDay": "06:00:00"},
                {"heatSetpoint": 16.0, "TimeOfDay": "09:00:00"},
                {"heatSetpoint": 20.0, "TimeOfDay": "20:00:00"},
            ],
        },
        {
            "DayOfWeek": 1,
            "Switchpoints": [
                {"heatSetpoint": 16.0, "TimeOfDay": "00:00:00"},
                {"heatSetpoint": 19.0, "TimeOfDay": "06:00:00"},
                {"heatSetpoint": 16.0, "TimeOfDay": "09:00:00"},
                {"heatSetpoint": 20.0, "TimeOfDay": "20:00:00"},
            ],
        },
        {
            "DayOfWeek": 2,
            "Switchpoints": [
                {"heatSetpoint": 16.0, "TimeOfDay": "00:00:00"},
                {"heatSetpoint": 19.0, "TimeOfDay": "06:00:00"},
                {"heatSetpoint": 16.0, "TimeOfDay": "09:00:00"},
                {"heatSetpoint": 20.0, "TimeOfDay": "20:00:00"},
            ],
        },
        {
            "DayOfWeek": 3,
            "Switchpoints": [
                {"heatSetpoint": 16.0, "TimeOfDay": "00:00:00"},
                {"heatSetpoint": 19.0, "TimeOfDay": "06:00:00"},
                {"heatSetpoint": 16.0, "TimeOfDay": "09:00:00"},
                {"heatSetpoint": 20.0, "TimeOfDay": "20:00:00"},
            ],
        },
        {
            "DayOfWeek": 4,
            "Switchpoints": [
                {"heatSetpoint": 16.0, "TimeOfDay": "00:00:00"},
                {"heatSetpoint": 19.0, "TimeOfDay": "06:00:00"},
                {"heatSetpoint": 16.0, "TimeOfDay": "09:00:00"},
                {"heatSetpoint": 20.0, "TimeOfDay": "20:00:00"},
            ],
        },
        {
            "DayOfWeek": 5,
            "Switchpoints": [
                {"heatSetpoint": 16.0, "TimeOfDay": "00:00:00"},
                {"heatSetpoint": 19.0, "TimeOfDay": "06:00:00"},
                {"heatSetpoint": 16.0, "TimeOfDay": "09:00:00"},
                {"heatSetpoint": 20.0, "TimeOfDay": "20:00:00"},
            ],
        },
        {
            "DayOfWeek": 6,
            "Switchpoints": [
                {"heatSetpoint": 16.0, "TimeOfDay": "00:00:00"},
                {"heatSetpoint": 19.0, "TimeOfDay": "06:00:00"},
                {"heatSetpoint": 16.0, "TimeOfDay": "09:00:00"},
                {"heatSetpoint": 20.0, "TimeOfDay": "20:00:00"},
            ],
        },
    ]
}
schedule["Basement Two"] = {
    "DailySchedules": [
        {
            "DayOfWeek": 0,
            "Switchpoints": [{"heatSetpoint": 16.0, "TimeOfDay": "06:00:00"}],
        },
        {
            "DayOfWeek": 1,
            "Switchpoints": [{"heatSetpoint": 16.0, "TimeOfDay": "06:00:00"}],
        },
        {
            "DayOfWeek": 2,
            "Switchpoints": [{"heatSetpoint": 16.0, "TimeOfDay": "06:00:00"}],
        },
        {
            "DayOfWeek": 3,
            "Switchpoints": [{"heatSetpoint": 16.0, "TimeOfDay": "06:00:00"}],
        },
        {
            "DayOfWeek": 4,
            "Switchpoints": [{"heatSetpoint": 16.0, "TimeOfDay": "06:00:00"}],
        },
        {
            "DayOfWeek": 5,
            "Switchpoints": [{"heatSetpoint": 16.0, "TimeOfDay": "06:00:00"}],
        },
        {
            "DayOfWeek": 6,
            "Switchpoints": [{"heatSetpoint": 16.0, "TimeOfDay": "06:00:00"}],
        },
    ]
}
schedule["Bathroom"] = {
    "DailySchedules": [
        {
            "DayOfWeek": 0,
            "Switchpoints": [
                {"heatSetpoint": 14.0, "TimeOfDay": "00:00:00"},
                {"heatSetpoint": 21.0, "TimeOfDay": "06:00:00"},
                {"heatSetpoint": 14.0, "TimeOfDay": "09:00:00"},
                {"heatSetpoint": 21.0, "TimeOfDay": "18:00:00"},
            ],
        },
        {
            "DayOfWeek": 1,
            "Switchpoints": [
                {"heatSetpoint": 14.0, "TimeOfDay": "00:00:00"},
                {"heatSetpoint": 21.0, "TimeOfDay": "06:00:00"},
                {"heatSetpoint": 14.0, "TimeOfDay": "09:00:00"},
                {"heatSetpoint": 21.0, "TimeOfDay": "18:00:00"},
            ],
        },
        {
            "DayOfWeek": 2,
            "Switchpoints": [
                {"heatSetpoint": 14.0, "TimeOfDay": "00:00:00"},
                {"heatSetpoint": 21.0, "TimeOfDay": "06:00:00"},
                {"heatSetpoint": 14.0, "TimeOfDay": "09:00:00"},
                {"heatSetpoint": 21.0, "TimeOfDay": "18:00:00"},
            ],
        },
        {
            "DayOfWeek": 3,
            "Switchpoints": [
                {"heatSetpoint": 14.0, "TimeOfDay": "00:00:00"},
                {"heatSetpoint": 21.0, "TimeOfDay": "06:00:00"},
                {"heatSetpoint": 14.0, "TimeOfDay": "09:00:00"},
                {"heatSetpoint": 21.0, "TimeOfDay": "18:00:00"},
            ],
        },
        {
            "DayOfWeek": 4,
            "Switchpoints": [
                {"heatSetpoint": 14.0, "TimeOfDay": "00:00:00"},
                {"heatSetpoint": 21.0, "TimeOfDay": "06:00:00"},
                {"heatSetpoint": 14.0, "TimeOfDay": "09:00:00"},
                {"heatSetpoint": 21.0, "TimeOfDay": "18:00:00"},
            ],
        },
        {
            "DayOfWeek": 5,
            "Switchpoints": [
                {"heatSetpoint": 14.0, "TimeOfDay": "00:00:00"},
                {"heatSetpoint": 19.0, "TimeOfDay": "06:00:00"},
                {"heatSetpoint": 21.0, "TimeOfDay": "18:00:00"},
            ],
        },
        {
            "DayOfWeek": 6,
            "Switchpoints": [
                {"heatSetpoint": 14.0, "TimeOfDay": "00:00:00"},
                {"heatSetpoint": 19.0, "TimeOfDay": "06:00:00"},
                {"heatSetpoint": 21.0, "TimeOfDay": "18:00:00"},
            ],
        },
    ]
}
schedule["Dressing Room"] = {
    "DailySchedules": [
        {
            "DayOfWeek": 0,
            "Switchpoints": [
                {"heatSetpoint": 16.0, "TimeOfDay": "00:00:00"},
                {"heatSetpoint": 19.0, "TimeOfDay": "06:00:00"},
                {"heatSetpoint": 16.0, "TimeOfDay": "09:00:00"},
            ],
        },
        {
            "DayOfWeek": 1,
            "Switchpoints": [
                {"heatSetpoint": 16.0, "TimeOfDay": "00:00:00"},
                {"heatSetpoint": 19.0, "TimeOfDay": "06:00:00"},
                {"heatSetpoint": 16.0, "TimeOfDay": "09:00:00"},
            ],
        },
        {
            "DayOfWeek": 2,
            "Switchpoints": [
                {"heatSetpoint": 16.0, "TimeOfDay": "00:00:00"},
                {"heatSetpoint": 19.0, "TimeOfDay": "06:00:00"},
                {"heatSetpoint": 16.0, "TimeOfDay": "09:00:00"},
            ],
        },
        {
            "DayOfWeek": 3,
            "Switchpoints": [
                {"heatSetpoint": 16.0, "TimeOfDay": "00:00:00"},
                {"heatSetpoint": 19.0, "TimeOfDay": "06:00:00"},
                {"heatSetpoint": 16.0, "TimeOfDay": "09:00:00"},
            ],
        },
        {
            "DayOfWeek": 4,
            "Switchpoints": [
                {"heatSetpoint": 16.0, "TimeOfDay": "00:00:00"},
                {"heatSetpoint": 19.0, "TimeOfDay": "06:00:00"},
                {"heatSetpoint": 16.0, "TimeOfDay": "09:00:00"},
            ],
        },
        {
            "DayOfWeek": 5,
            "Switchpoints": [
                {"heatSetpoint": 16.0, "TimeOfDay": "00:00:00"},
                {"heatSetpoint": 19.0, "TimeOfDay": "06:00:00"},
                {"heatSetpoint": 16.0, "TimeOfDay": "09:00:00"},
            ],
        },
        {
            "DayOfWeek": 6,
            "Switchpoints": [
                {"heatSetpoint": 16.0, "TimeOfDay": "00:00:00"},
                {"heatSetpoint": 19.0, "TimeOfDay": "06:00:00"},
                {"heatSetpoint": 16.0, "TimeOfDay": "09:00:00"},
            ],
        },
    ]
}
schedule["Sitting Room"] = {
    "DailySchedules": [
        {
            "DayOfWeek": 0,
            "Switchpoints": [
                {"heatSetpoint": 19.0, "TimeOfDay": "06:30:00"},
                {"heatSetpoint": 14.0, "TimeOfDay": "09:00:00"},
                {"heatSetpoint": 21.0, "TimeOfDay": "17:30:00"},
                {"heatSetpoint": 14.0, "TimeOfDay": "23:00:00"},
            ],
        },
        {
            "DayOfWeek": 1,
            "Switchpoints": [
                {"heatSetpoint": 19.0, "TimeOfDay": "06:30:00"},
                {"heatSetpoint": 14.0, "TimeOfDay": "09:00:00"},
                {"heatSetpoint": 21.0, "TimeOfDay": "17:30:00"},
                {"heatSetpoint": 14.0, "TimeOfDay": "23:00:00"},
            ],
        },
        {
            "DayOfWeek": 2,
            "Switchpoints": [
                {"heatSetpoint": 19.0, "TimeOfDay": "06:30:00"},
                {"heatSetpoint": 14.0, "TimeOfDay": "09:00:00"},
                {"heatSetpoint": 21.0, "TimeOfDay": "17:30:00"},
                {"heatSetpoint": 14.0, "TimeOfDay": "23:00:00"},
            ],
        },
        {
            "DayOfWeek": 3,
            "Switchpoints": [
                {"heatSetpoint": 19.0, "TimeOfDay": "06:30:00"},
                {"heatSetpoint": 14.0, "TimeOfDay": "09:00:00"},
                {"heatSetpoint": 21.0, "TimeOfDay": "17:30:00"},
                {"heatSetpoint": 14.0, "TimeOfDay": "23:00:00"},
            ],
        },
        {
            "DayOfWeek": 4,
            "Switchpoints": [
                {"heatSetpoint": 19.0, "TimeOfDay": "06:30:00"},
                {"heatSetpoint": 14.0, "TimeOfDay": "09:00:00"},
                {"heatSetpoint": 21.0, "TimeOfDay": "17:30:00"},
                {"heatSetpoint": 14.0, "TimeOfDay": "23:00:00"},
            ],
        },
        {
            "DayOfWeek": 5,
            "Switchpoints": [
                {"heatSetpoint": 19.0, "TimeOfDay": "06:30:00"},
                {"heatSetpoint": 21.0, "TimeOfDay": "17:30:00"},
                {"heatSetpoint": 14.0, "TimeOfDay": "23:00:00"},
            ],
        },
        {
            "DayOfWeek": 6,
            "Switchpoints": [
                {"heatSetpoint": 19.0, "TimeOfDay": "06:30:00"},
                {"heatSetpoint": 21.0, "TimeOfDay": "17:30:00"},
                {"heatSetpoint": 14.0, "TimeOfDay": "23:00:00"},
            ],
        },
    ]
}
schedule["Hall"] = {
    "DailySchedules": [
        {
            "DayOfWeek": 0,
            "Switchpoints": [{"heatSetpoint": 16.0, "TimeOfDay": "00:00:00"}],
        },
        {
            "DayOfWeek": 1,
            "Switchpoints": [{"heatSetpoint": 16.0, "TimeOfDay": "00:00:00"}],
        },
        {
            "DayOfWeek": 2,
            "Switchpoints": [{"heatSetpoint": 16.0, "TimeOfDay": "00:00:00"}],
        },
        {
            "DayOfWeek": 3,
            "Switchpoints": [
                {"heatSetpoint": 16.0, "TimeOfDay": "00:00:00"},
                {"heatSetpoint": 21.0, "TimeOfDay": "06:30:00"},
                {"heatSetpoint": 18.0, "TimeOfDay": "08:00:00"},
                {"heatSetpoint": 21.0, "TimeOfDay": "18:00:00"},
                {"heatSetpoint": 16.0, "TimeOfDay": "22:30:00"},
            ],
        },
        {
            "DayOfWeek": 4,
            "Switchpoints": [{"heatSetpoint": 16.0, "TimeOfDay": "00:00:00"}],
        },
        {
            "DayOfWeek": 5,
            "Switchpoints": [{"heatSetpoint": 16.0, "TimeOfDay": "00:00:00"}],
        },
    ]
}



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
"""Mocked vendor RESTful API via a hacked aiohttp."""
from __future__ import annotations

from .aiohttp import ClientSession  # noqa: F401
from .vendor import MockedServer  # noqa: F401



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
"""evohomeasync2 - Schema for RESTful API Config JSON."""
from __future__ import annotations

from .const import (
    REGEX_DHW_ID,
    REGEX_SYSTEM_ID,
    REGEX_ZONE_ID,
    SZ_ALLOWED_MODES,
    SZ_ALLOWED_SETPOINT_MODES,
    SZ_ALLOWED_STATES,
    SZ_ALLOWED_SYSTEM_MODES,
    SZ_CAN_BE_PERMANENT,
    SZ_CAN_BE_TEMPORARY,
    SZ_CAN_CONTROL_COOL,
    SZ_CAN_CONTROL_HEAT,
    SZ_CITY,
    SZ_CURRENT_OFFSET_MINUTES,
    SZ_COUNTRY,
    SZ_CRC,
    SZ_DHW,
    SZ_DHW_ID,
    SZ_DHW_STATE_CAPABILITIES_RESPONSE,
    SZ_FIRSTNAME,
    SZ_DISPLAY_NAME,
    SZ_GATEWAY_ID,
    SZ_GATEWAY_INFO,
    SZ_GATEWAYS,
    SZ_IS_WI_FI,
    SZ_LASTNAME,
    SZ_LOCATION_ID,
    SZ_LOCATION_INFO,
    SZ_LOCATION_TYPE,
    SZ_LOCATION_OWNER,
    SZ_MAC,
    SZ_MAX_COOL_SETPOINT,
    SZ_MAX_DURATION,
    SZ_MAX_HEAT_SETPOINT,
    SZ_MAX_SWITCHPOINTS_PER_DAY,
    SZ_MIN_COOL_SETPOINT,
    SZ_MIN_HEAT_SETPOINT,
    SZ_MIN_SWITCHPOINTS_PER_DAY,
    SZ_MODEL_TYPE,
    SZ_NAME,
    SZ_OFFSET_MINUTES,
    SZ_POSTCODE,
    SZ_SCHEDULE_CAPABILITIES,
    SZ_SCHEDULE_CAPABILITIES_RESPONSE,
    SZ_SETPOINT_CAPABILITIES,
    SZ_SETPOINT_VALUE_RESOLUTION,
    SZ_STREET_ADDRESS,
    SZ_SUPPORTS_DAYLIGHT_SAVING,
    SZ_SYSTEM_ID,
    SZ_SYSTEM_MODE,
    SZ_TEMPERATURE_CONTROL_SYSTEMS,
    SZ_TIME_ZONE,
    SZ_TIME_ZONE_ID,
    SZ_TIMING_MODE,
    SZ_TIMING_RESOLUTION,
    SZ_USE_DAYLIGHT_SAVE_SWITCHING,
    SZ_USER_ID,
    SZ_USERNAME,
    SZ_VACATION_HOLD_CAPABILITIES,
    SZ_VALUE_RESOLUTION,
    SZ_ZONE_ID,
    SZ_ZONE_TYPE,
    SZ_ZONES,
)
from .const import DhwState, SystemMode, TcsModelType, ZoneMode, ZoneModelType, ZoneType
from .const import obfuscate as _obfuscate
from .helpers import vol  # voluptuous

SZ_DURATION = "Duration"
SZ_PERIOD = "Period"


SCH_SYSTEM_MODE_PERM = vol.Schema(
    {
        vol.Required(SZ_SYSTEM_MODE): vol.Any(
            str(SystemMode.AUTO),
            str(SystemMode.AUTO_WITH_RESET),
            str(SystemMode.HEATING_OFF),
            str(SystemMode.OFF),  # not evohome
            str(SystemMode.HEAT),  # not evohome
        ),
        vol.Required(SZ_CAN_BE_PERMANENT): True,
        vol.Required(SZ_CAN_BE_TEMPORARY): False,
    },
    extra=vol.PREVENT_EXTRA,
)  # TODO: does this apply to DHW?

SCH_SYSTEM_MODE_TEMP = vol.Schema(
    {
        vol.Required(SZ_SYSTEM_MODE): vol.Any(
            str(SystemMode.AUTO_WITH_ECO),
            str(SystemMode.AWAY),
            str(SystemMode.CUSTOM),
            str(SystemMode.DAY_OFF),
        ),
        vol.Required(SZ_CAN_BE_PERMANENT): True,
        vol.Required(SZ_CAN_BE_TEMPORARY): True,
        vol.Required(SZ_MAX_DURATION): str,  # "99.00:00:00"
        vol.Required(SZ_TIMING_RESOLUTION): str,  # "1.00:00:00"
        vol.Required(SZ_TIMING_MODE): vol.Any(SZ_DURATION, SZ_PERIOD),
    },
    extra=vol.PREVENT_EXTRA,
)

SCH_ALLOWED_SYSTEM_MODE = vol.Any(SCH_SYSTEM_MODE_PERM, SCH_SYSTEM_MODE_TEMP)


SCH_DHW_STATE_CAPABILITIES_RESPONSE = vol.Schema(
    {
        vol.Required(SZ_ALLOWED_STATES): list(m.value for m in DhwState),
        vol.Required(SZ_ALLOWED_MODES): list(m.value for m in ZoneMode),
        vol.Required(SZ_MAX_DURATION): str,
        vol.Required(SZ_TIMING_RESOLUTION): vol.Datetime(format="00:%M:00"),
    },
    extra=vol.PREVENT_EXTRA,
)

SCH_SCHEDULE_CAPABILITIES_RESPONSE = vol.Schema(
    {
        vol.Required(SZ_MAX_SWITCHPOINTS_PER_DAY): int,  # 6
        vol.Required(SZ_MIN_SWITCHPOINTS_PER_DAY): int,  # 1
        vol.Required(SZ_TIMING_RESOLUTION): vol.Datetime(
            format="00:%M:00"
        ),  # "00:10:00"
    },
    extra=vol.PREVENT_EXTRA,
)

SCH_DHW = vol.Schema(
    {
        vol.Required(SZ_DHW_ID): vol.Match(REGEX_DHW_ID),
        vol.Required(
            SZ_DHW_STATE_CAPABILITIES_RESPONSE
        ): SCH_DHW_STATE_CAPABILITIES_RESPONSE,
        vol.Required(
            SZ_SCHEDULE_CAPABILITIES_RESPONSE
        ): SCH_SCHEDULE_CAPABILITIES_RESPONSE,
    },
    extra=vol.PREVENT_EXTRA,
)

SCH_SETPOINT_CAPABILITIES = vol.Schema(
    {
        vol.Required(SZ_CAN_CONTROL_HEAT): bool,
        vol.Required(SZ_MAX_HEAT_SETPOINT): float,  # 35.0
        vol.Required(SZ_MIN_HEAT_SETPOINT): float,  # 5.0
        vol.Required(SZ_CAN_CONTROL_COOL): bool,
        vol.Optional(SZ_MAX_COOL_SETPOINT): float,  #
        vol.Optional(SZ_MIN_COOL_SETPOINT): float,  #
        vol.Required(SZ_ALLOWED_SETPOINT_MODES): list(m.value for m in ZoneMode),
        vol.Required(SZ_VALUE_RESOLUTION): float,  # 0.5
        vol.Required(SZ_MAX_DURATION): str,  # "1.00:00:00"
        vol.Required(SZ_TIMING_RESOLUTION): vol.Datetime(
            format="00:%M:00"
        ),  # "00:10:00"
        vol.Optional(SZ_VACATION_HOLD_CAPABILITIES): dict,  # TODO: non-evohome
        # vol.Optional(SZ_ALLOWED_FAN_MODES): dict,  # TODO: non-evohome
    },
    extra=vol.PREVENT_EXTRA,
)

SCH_SCHEDULE_CAPABILITIES = SCH_SCHEDULE_CAPABILITIES_RESPONSE.extend(
    {
        vol.Required(SZ_SETPOINT_VALUE_RESOLUTION): float,
    },
    extra=vol.PREVENT_EXTRA,
)

SCH_ZONE = vol.Schema(
    {
        vol.Required(SZ_ZONE_ID): vol.Match(REGEX_ZONE_ID),
        vol.Required(SZ_MODEL_TYPE): vol.Any(*(m.value for m in ZoneModelType)),
        vol.Required(SZ_NAME): str,
        vol.Required(SZ_SETPOINT_CAPABILITIES): SCH_SETPOINT_CAPABILITIES,
        vol.Required(SZ_SCHEDULE_CAPABILITIES): SCH_SCHEDULE_CAPABILITIES,
        vol.Required(SZ_ZONE_TYPE): vol.Any(*(m.value for m in ZoneType)),
    },
    extra=vol.PREVENT_EXTRA,
)

SCH_TEMPERATURE_CONTROL_SYSTEM = vol.Schema(
    {
        vol.Required(SZ_SYSTEM_ID): vol.Match(REGEX_SYSTEM_ID),
        vol.Required(SZ_MODEL_TYPE): vol.Any(*(m.value for m in TcsModelType)),
        vol.Required(SZ_ALLOWED_SYSTEM_MODES): [SCH_ALLOWED_SYSTEM_MODE],
        vol.Required(SZ_ZONES): vol.All([SCH_ZONE], vol.Length(min=1, max=12)),
        vol.Optional(SZ_DHW): SCH_DHW,
    },
    extra=vol.PREVENT_EXTRA,
)

SCH_GATEWAY_INFO = vol.Schema(
    {
        vol.Required(SZ_GATEWAY_ID): str,
        vol.Required(SZ_MAC): str,
        vol.Required(SZ_CRC): vol.All(str, _obfuscate),
        vol.Required(SZ_IS_WI_FI): bool,
    },
    extra=vol.PREVENT_EXTRA,
)

SCH_GATEWAY = vol.Schema(
    {
        vol.Required(SZ_GATEWAY_INFO): SCH_GATEWAY_INFO,
        vol.Required(SZ_TEMPERATURE_CONTROL_SYSTEMS): [SCH_TEMPERATURE_CONTROL_SYSTEM],
    },
    extra=vol.PREVENT_EXTRA,
)

SCH_TIME_ZONE = vol.Schema(
    {
        vol.Required(SZ_TIME_ZONE_ID): str,
        vol.Required(SZ_DISPLAY_NAME): str,
        vol.Required(SZ_OFFSET_MINUTES): int,
        vol.Required(SZ_CURRENT_OFFSET_MINUTES): int,
        vol.Required(SZ_SUPPORTS_DAYLIGHT_SAVING): bool,
    },
    extra=vol.PREVENT_EXTRA,
)

SCH_LOCATION_OWNER = vol.Schema(
    {
        vol.Required(SZ_USER_ID): str,
        vol.Required(SZ_USERNAME): vol.All(vol.Email(), _obfuscate),
        vol.Required(SZ_FIRSTNAME): str,
        vol.Required(SZ_LASTNAME): vol.All(str, _obfuscate),
    },
    extra=vol.PREVENT_EXTRA,
)

SCH_LOCATION_INFO = vol.Schema(
    {
        vol.Required(SZ_LOCATION_ID): str,
        vol.Required(SZ_NAME): str,  # e.g. "My Home"
        vol.Required(SZ_STREET_ADDRESS): vol.All(str, _obfuscate),
        vol.Required(SZ_CITY): vol.All(str, _obfuscate),
        vol.Required(SZ_COUNTRY): str,
        vol.Required(SZ_POSTCODE): vol.All(str, _obfuscate),
        vol.Required(SZ_LOCATION_TYPE): str,  # "Residential"
        vol.Required(SZ_USE_DAYLIGHT_SAVE_SWITCHING): bool,
        vol.Required(SZ_TIME_ZONE): SCH_TIME_ZONE,
        vol.Required(SZ_LOCATION_OWNER): SCH_LOCATION_OWNER,
    },
    extra=vol.PREVENT_EXTRA,
)

# /location/{location_id}/installationInfo?includeTemperatureControlSystems=True
SCH_LOCATION_INSTALLATION_INFO = vol.Schema(
    {
        vol.Required(SZ_LOCATION_INFO): SCH_LOCATION_INFO,
        vol.Required(SZ_GATEWAYS): [SCH_GATEWAY],
    },
    extra=vol.PREVENT_EXTRA,
)

# /location/installationInfo?userId={user_id}&includeTemperatureControlSystems=True
SCH_USER_LOCATIONS_INSTALLATION_INFO = vol.Schema(
    [SCH_LOCATION_INSTALLATION_INFO],
    extra=vol.PREVENT_EXTRA,
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
"""Provides handling of TCC DHW zones."""

# TODO: add set_mode() for non-evohome modes

from __future__ import annotations

from datetime import datetime as dt
from typing import TYPE_CHECKING, Final, NoReturn

from .const import API_STRFTIME
from .exceptions import DeprecationError, InvalidSchema
from .schema import SCH_DHW_STATUS
from .schema.const import DhwState, ZoneMode
from .schema.const import (
    SZ_ALLOWED_MODES,
    SZ_DHW_ID,
    SZ_DHW_STATE_CAPABILITIES_RESPONSE,
    SZ_DOMESTIC_HOT_WATER,
    SZ_MODE,
    SZ_SCHEDULE_CAPABILITIES_RESPONSE,
    SZ_STATE,
    SZ_STATE_STATUS,
    SZ_UNTIL_TIME,
)
from .schema.schedule import SCH_GET_SCHEDULE_DHW, SCH_PUT_SCHEDULE_DHW
from .zone import _ZoneBase


if TYPE_CHECKING:
    from . import ControlSystem
    from .schema import _DhwIdT, _EvoDictT, _EvoListT


class HotWaterDeprecated:
    """Deprecated attributes and methods removed from the evohome-client namespace."""

    @property
    def zoneId(self) -> NoReturn:
        raise DeprecationError("HotWater.zoneId is deprecated, use .dhwId (or ._id)")

    async def get_dhw_state(self, *args, **kwargs) -> NoReturn:
        raise DeprecationError(
            "HotWater.get_dhw_state() is deprecated, use Location.refresh_status()"
        )

    async def set_dhw_auto(self, *args, **kwargs) -> NoReturn:
        raise DeprecationError(
            "HotWater.set_dhw_auto() is deprecated, use .reset_mode()"
        )

    async def set_dhw_off(self, *args, **kwargs) -> NoReturn:
        raise DeprecationError("HotWater.set_dhw_off() is deprecated, use .set_off()")

    async def set_dhw_on(self, *args, **kwargs) -> NoReturn:
        raise DeprecationError("HotWater.set_dhw_on() is deprecated, use .set_on()")


class HotWater(HotWaterDeprecated, _ZoneBase):
    """Instance of a TCS's DHW zone (domesticHotWater)."""

    STATUS_SCHEMA = SCH_DHW_STATUS
    TYPE: Final[str] = SZ_DOMESTIC_HOT_WATER  # type: ignore[misc]

    SCH_SCHEDULE_GET = SCH_GET_SCHEDULE_DHW
    SCH_SCHEDULE_PUT = SCH_PUT_SCHEDULE_DHW

    def __init__(self, tcs: ControlSystem, config: _EvoDictT) -> None:
        super().__init__(tcs)

        self._config: Final[_EvoDictT] = config

        try:
            assert self.dhwId, "Invalid config dict"
        except AssertionError as exc:
            raise InvalidSchema(str(exc))
        self._id = self.dhwId

    @property
    def dhwId(self) -> _DhwIdT:
        return self._config[SZ_DHW_ID]

    @property
    def dhwStateCapabilitiesResponse(self) -> _EvoDictT:
        return self._config[SZ_DHW_STATE_CAPABILITIES_RESPONSE]

    @property
    def scheduleCapabilitiesResponse(self) -> _EvoDictT:
        return self._config[SZ_SCHEDULE_CAPABILITIES_RESPONSE]

    @property  # for convenience (is not a top-level config attribute)
    def allowedModes(self) -> _EvoListT:
        return self._config[SZ_SCHEDULE_CAPABILITIES_RESPONSE][SZ_ALLOWED_MODES]

    @property
    def name(self) -> str:
        return "Domestic Hot Water"

    @property
    def stateStatus(self) -> _EvoDictT | None:
        return self._status.get(SZ_STATE_STATUS)

    async def _set_mode(self, mode: dict) -> None:
        """Set the DHW mode (state)."""
        _ = await self._broker.put(f"{self.TYPE}/{self._id}/state", json=mode)

    async def reset_mode(self) -> None:
        """Cancel any override and allow the DHW to follow its schedule."""

        mode: dict[str, str | None] = {  # NOTE: SZ_STATE was previously ""
            SZ_MODE: ZoneMode.FOLLOW_SCHEDULE,
            SZ_STATE: None,
            SZ_UNTIL_TIME: None,
        }

        await self._set_mode(mode)

    async def set_on(self, /, *, until: dt | None = None) -> None:
        """Set the DHW on until a given time, or permanently."""

        mode: dict[str, str | None]

        if until is None:
            mode = {
                SZ_MODE: ZoneMode.PERMANENT_OVERRIDE,
                SZ_STATE: DhwState.ON,
                SZ_UNTIL_TIME: None,
            }
        else:
            mode = {
                SZ_MODE: ZoneMode.TEMPORARY_OVERRIDE,
                SZ_STATE: DhwState.ON,
                SZ_UNTIL_TIME: until.strftime(API_STRFTIME),
            }

        await self._set_mode(mode)

    async def set_off(self, /, *, until: dt | None = None) -> None:
        """Set the DHW off until a given time, or permanently."""

        mode: dict[str, str | None]

        if until is None:
            mode = {
                SZ_MODE: ZoneMode.PERMANENT_OVERRIDE,
                SZ_STATE: DhwState.OFF,
                SZ_UNTIL_TIME: None,
            }
        else:
            mode = {
                SZ_MODE: ZoneMode.TEMPORARY_OVERRIDE,
                SZ_STATE: DhwState.OFF,
                SZ_UNTIL_TIME: until.strftime(API_STRFTIME),
            }

        await self._set_mode(mode)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
"""evohome-async - validate the handling of vendor APIs (URLs)."""
from __future__ import annotations


from http import HTTPMethod, HTTPStatus
import pytest
import pytest_asyncio

import evohomeasync as evo
from evohomeasync.client import URL_HOST
from evohomeasync2.schema import vol  # FIXME: need v1 schemas

from . import _DEBUG_USE_REAL_AIOHTTP, _DISABLE_STRICT_ASSERTS
from .helpers import aiohttp  # aiohttp may be mocked
from .helpers import user_credentials as _user_credentials
from .helpers import client_session as _client_session


URL_BASE = f"{URL_HOST}/WebAPI/api"


_global_user_data: dict | None = None


@pytest.fixture()
def user_credentials():
    return _user_credentials()


@pytest_asyncio.fixture
async def session():
    client_session = _client_session()
    try:
        yield client_session
    finally:
        await client_session.close()


@pytest.fixture(autouse=True)
def patches_for_tests(monkeypatch: pytest.MonkeyPatch):
    monkeypatch.setattr("evohomeasync.client.aiohttp", aiohttp)


async def instantiate_client(
    username: str,
    password: str,
    session: aiohttp.ClientSession | None = None,
):
    """Instantiate a client, and logon to the vendor API."""

    global _global_user_data

    # refresh_token, access_token, access_token_expires = _global_user_data

    # Instantiation, NOTE: No API calls invoked during instantiation
    client = evo.EvohomeClient(
        username,
        password,
        session=session,
        user_data=_global_user_data,
    )

    # Authentication
    await client._populate_user_data()
    _global_user_data = client.user_data

    return client


async def should_work(
    client: evo.EvohomeClient,
    method: HTTPMethod,
    url: str,
    json: dict | None = None,
    content_type: str | None = "application/json",
    schema: vol.Schema | None = None,
) -> dict | list:
    """Make a request that is expected to succeed."""

    response: aiohttp.ClientResponse

    response = await client._do_request(method, f"{URL_BASE}/{url}", data=json)

    response.raise_for_status()

    assert True or response.content_type == content_type

    if response.content_type == "application/json":
        content = await response.json()
    else:
        content = await response.text()

    if schema:
        return schema(content)
    return content


async def should_fail(
    client: evo.EvohomeClient,
    method: HTTPMethod,
    url: str,
    json: dict | None = None,
    content_type: str | None = "application/json",
    status: HTTPStatus | None = None,
) -> aiohttp.ClientResponse:
    """Make a request that is expected to fail."""

    response: aiohttp.ClientResponse

    response = await client._do_request(method, f"{URL_BASE}/{url}", data=json)

    try:
        response.raise_for_status()
    except aiohttp.ClientResponseError as exc:
        assert exc.status == status
    else:
        assert False, response.status

    if _DISABLE_STRICT_ASSERTS:
        return response

    assert True or response.content_type == content_type

    if response.content_type == "application/json":
        content = await response.json()
    else:
        content = await response.text()

    if isinstance(content, dict):
        assert True or "message" in content
    elif isinstance(content, list):
        assert True or "message" in content[0]

    return content


async def _test_url_locations(
    username: str, password: str, session: aiohttp.ClientSession | None = None
) -> None:
    client = await instantiate_client(username, password, session=session)

    client.headers["sessionId"] = client.user_data["sessionId"]
    user_id: int = client.user_data["userInfo"]["userID"]

    url = f"/locations?userId={user_id}&allData=True"
    _ = await should_work(client, HTTPMethod.GET, url)
    _ = await should_fail(
        client, HTTPMethod.PUT, url, status=HTTPStatus.METHOD_NOT_ALLOWED
    )

    url = f"/locations?userId={user_id}"
    _ = await should_work(client, HTTPMethod.GET, url)

    url = "/locations?userId=123456"
    _ = await should_fail(client, HTTPMethod.GET, url)

    url = "/locations?userId='123456'"
    _ = await should_fail(client, HTTPMethod.GET, url)

    url = "xxxxxxx"  # NOTE: is a general test, and not a test specific to this URL
    _ = await should_fail(
        client,
        HTTPMethod.GET,
        url,
        status=HTTPStatus.NOT_FOUND,
        content_type="text/html",  # exception to usual content-type
    )


async def _test_client_apis(
    username: str,
    password: str,
    session: aiohttp.ClientSession | None = None,
):
    """Instantiate a client, and logon to the vendor API."""

    global _global_user_data

    # Instantiation, NOTE: No API calls invoked during instantiation
    client = evo.EvohomeClient(
        username,
        password,
        session=session,
        user_data=_global_user_data,
    )

    user_data = await client._populate_user_data()
    assert user_data  # aka client.user_data

    _global_user_data = client.user_data

    await client._populate_full_data()
    assert client.full_data

    temps = await client.temperatures()
    assert temps


@pytest.mark.asyncio
async def _test_locations(
    user_credentials: tuple[str, str], session: aiohttp.ClientSession
) -> None:
    """Test /locations"""

    if not _DEBUG_USE_REAL_AIOHTTP:
        pytest.skip("Mocked server not implemented")

    try:
        await _test_url_locations(*user_credentials, session=session)
    except evo.AuthenticationFailed:
        pytest.skip("Unable to authenticate")


@pytest.mark.asyncio
async def test_client_apis(
    user_credentials: tuple[str, str], session: aiohttp.ClientSession
) -> None:
    """Test _populate_user_data() & _populate_full_data()"""

    if not _DEBUG_USE_REAL_AIOHTTP:
        pytest.skip("Mocked server not implemented")

    try:
        await _test_client_apis(*user_credentials, session=session)
    except evo.AuthenticationFailed:
        pytest.skip("Unable to authenticate")


USER_DATA = {
    "sessionId": "BE5F40A6-1234-1234-1234-A708947D638B",
    "userInfo": {
        "userID": 1234567,
        "username": "null@gmail.com",
        "firstname": "David",
        "lastname": "Smith",
        "streetAddress": "1 Main Street",
        "city": "London",
        "zipcode": "NW1 1AA",
        "country": "GB",
        "telephone": "",
        "userLanguage": "en-GB",
        "isActivated": True,
        "deviceCount": 0,
        "tenantID": 5,
        "securityQuestion1": "NotUsed",
        "securityQuestion2": "NotUsed",
        "securityQuestion3": "NotUsed",
        "latestEulaAccepted": False,
    },
}

FULL_DATA = {
    "locationID": 2738909,
    "name": "My Home",
    "streetAddress": "1 Main Street",
    "city": "London",
    "state": "",
    "country": "GB",
    "zipcode": "NW1 1AA",
    "type": "Residential",
    "hasStation": True,
    "devices": [
        {
            "gatewayId": 2499896,
            "deviceID": 3933910,
            "thermostatModelType": "DOMESTIC_HOT_WATER",
            "deviceType": 128,
            "name": "",
            "scheduleCapable": False,
            "holdUntilCapable": False,
            "thermostat": {
                "units": "Celsius",
                "indoorTemperature": 22.77,
                "outdoorTemperature": 128.0,
                "outdoorTemperatureAvailable": False,
                "outdoorHumidity": 128.0,
                "outdootHumidityAvailable": False,
                "indoorHumidity": 128.0,
                "indoorTemperatureStatus": "Measured",
                "indoorHumidityStatus": "NotAvailable",
                "outdoorTemperatureStatus": "NotAvailable",
                "outdoorHumidityStatus": "NotAvailable",
                "isCommercial": False,
                "allowedModes": ["DHWOn", "DHWOff"],
                "deadband": 0.0,
                "minHeatSetpoint": 5.0,
                "maxHeatSetpoint": 30.0,
                "minCoolSetpoint": 50.0,
                "maxCoolSetpoint": 90.0,
                "changeableValues": {"mode": "DHWOff", "status": "Scheduled"},
                "scheduleCapable": False,
                "vacationHoldChangeable": False,
                "vacationHoldCancelable": False,
                "scheduleHeatSp": 0.0,
                "scheduleCoolSp": 0.0,
            },
            "alertSettings": {
                "deviceID": 3933910,
                "tempHigherThanActive": True,
                "tempHigherThan": 30.0,
                "tempHigherThanMinutes": 0,
                "tempLowerThanActive": True,
                "tempLowerThan": 5.0,
                "tempLowerThanMinutes": 0,
                "faultConditionExistsActive": False,
                "faultConditionExistsHours": 0,
                "normalConditionsActive": True,
                "communicationLostActive": False,
                "communicationLostHours": 0,
                "communicationFailureActive": True,
                "communicationFailureMinutes": 15,
                "deviceLostActive": False,
                "deviceLostHours": 0,
            },
            "isUpgrading": False,
            "isAlive": True,
            "thermostatVersion": "02.00.19.33",
            "macID": "00D02DEE4E56",
            "locationID": 2738909,
            "domainID": 20054,
            "instance": 250,
        },
        {
            "gatewayId": 2499896,
            "deviceID": 3432579,
            "thermostatModelType": "EMEA_ZONE",
            "deviceType": 128,
            "name": "Bathroom Dn",
            "scheduleCapable": False,
            "holdUntilCapable": False,
            "thermostat": {
                "units": "Celsius",
                "indoorTemperature": 20.79,
                "outdoorTemperature": 128.0,
                "outdoorTemperatureAvailable": False,
                "outdoorHumidity": 128.0,
                "outdootHumidityAvailable": False,
                "indoorHumidity": 128.0,
                "indoorTemperatureStatus": "Measured",
                "indoorHumidityStatus": "NotAvailable",
                "outdoorTemperatureStatus": "NotAvailable",
                "outdoorHumidityStatus": "NotAvailable",
                "isCommercial": False,
                "allowedModes": ["Heat", "Off"],
                "deadband": 0.0,
                "minHeatSetpoint": 5.0,
                "maxHeatSetpoint": 35.0,
                "minCoolSetpoint": 50.0,
                "maxCoolSetpoint": 90.0,
                "changeableValues": {
                    "mode": "Off",
                    "heatSetpoint": {"value": 15.0, "status": "Scheduled"},
                    "vacationHoldDays": 0,
                },
                "scheduleCapable": False,
                "vacationHoldChangeable": False,
                "vacationHoldCancelable": False,
                "scheduleHeatSp": 0.0,
                "scheduleCoolSp": 0.0,
            },
            "alertSettings": {
                "deviceID": 3432579,
                "tempHigherThanActive": True,
                "tempHigherThan": 30.0,
                "tempHigherThanMinutes": 0,
                "tempLowerThanActive": True,
                "tempLowerThan": 5.0,
                "tempLowerThanMinutes": 0,
                "faultConditionExistsActive": False,
                "faultConditionExistsHours": 0,
                "normalConditionsActive": True,
                "communicationLostActive": False,
                "communicationLostHours": 0,
                "communicationFailureActive": True,
                "communicationFailureMinutes": 15,
                "deviceLostActive": False,
                "deviceLostHours": 0,
            },
            "isUpgrading": False,
            "isAlive": True,
            "thermostatVersion": "02.00.19.33",
            "macID": "00D02DEE4E56",
            "locationID": 2738909,
            "domainID": 20054,
            "instance": 4,
        },
        {
            "gatewayId": 2499896,
            "deviceID": 3449740,
            "thermostatModelType": "EMEA_ZONE",
            "deviceType": 128,
            "name": "Bathroom Up",
            "scheduleCapable": False,
            "holdUntilCapable": False,
            "thermostat": {
                "units": "Celsius",
                "indoorTemperature": 20.26,
                "outdoorTemperature": 128.0,
                "outdoorTemperatureAvailable": False,
                "outdoorHumidity": 128.0,
                "outdootHumidityAvailable": False,
                "indoorHumidity": 128.0,
                "indoorTemperatureStatus": "Measured",
                "indoorHumidityStatus": "NotAvailable",
                "outdoorTemperatureStatus": "NotAvailable",
                "outdoorHumidityStatus": "NotAvailable",
                "isCommercial": False,
                "allowedModes": ["Heat", "Off"],
                "deadband": 0.0,
                "minHeatSetpoint": 5.0,
                "maxHeatSetpoint": 35.0,
                "minCoolSetpoint": 50.0,
                "maxCoolSetpoint": 90.0,
                "changeableValues": {
                    "mode": "Off",
                    "heatSetpoint": {"value": 19.0, "status": "Scheduled"},
                    "vacationHoldDays": 0,
                },
                "scheduleCapable": False,
                "vacationHoldChangeable": False,
                "vacationHoldCancelable": False,
                "scheduleHeatSp": 0.0,
                "scheduleCoolSp": 0.0,
            },
            "alertSettings": {
                "deviceID": 3449740,
                "tempHigherThanActive": True,
                "tempHigherThan": 30.0,
                "tempHigherThanMinutes": 0,
                "tempLowerThanActive": True,
                "tempLowerThan": 5.0,
                "tempLowerThanMinutes": 0,
                "faultConditionExistsActive": False,
                "faultConditionExistsHours": 0,
                "normalConditionsActive": True,
                "communicationLostActive": False,
                "communicationLostHours": 0,
                "communicationFailureActive": True,
                "communicationFailureMinutes": 15,
                "deviceLostActive": False,
                "deviceLostHours": 0,
            },
            "isUpgrading": False,
            "isAlive": True,
            "thermostatVersion": "02.00.19.33",
            "macID": "00D02DEE4E56",
            "locationID": 2738909,
            "domainID": 20054,
            "instance": 7,
        },
        {
            "gatewayId": 2499896,
            "deviceID": 3432521,
            "thermostatModelType": "EMEA_ZONE",
            "deviceType": 128,
            "name": "Dead Zone",
            "scheduleCapable": False,
            "holdUntilCapable": False,
            "thermostat": {
                "units": "Celsius",
                "indoorTemperature": 128.0,
                "outdoorTemperature": 128.0,
                "outdoorTemperatureAvailable": False,
                "outdoorHumidity": 128.0,
                "outdootHumidityAvailable": False,
                "indoorHumidity": 128.0,
                "indoorTemperatureStatus": "NotAvailable",
                "indoorHumidityStatus": "NotAvailable",
                "outdoorTemperatureStatus": "NotAvailable",
                "outdoorHumidityStatus": "NotAvailable",
                "isCommercial": False,
                "allowedModes": ["Heat", "Off"],
                "deadband": 0.0,
                "minHeatSetpoint": 5.0,
                "maxHeatSetpoint": 35.0,
                "minCoolSetpoint": 50.0,
                "maxCoolSetpoint": 90.0,
                "changeableValues": {
                    "mode": "Off",
                    "heatSetpoint": {"value": 5.0, "status": "Scheduled"},
                    "vacationHoldDays": 0,
                },
                "scheduleCapable": False,
                "vacationHoldChangeable": False,
                "vacationHoldCancelable": False,
                "scheduleHeatSp": 0.0,
                "scheduleCoolSp": 0.0,
            },
            "alertSettings": {
                "deviceID": 3432521,
                "tempHigherThanActive": True,
                "tempHigherThan": 30.0,
                "tempHigherThanMinutes": 0,
                "tempLowerThanActive": True,
                "tempLowerThan": 5.0,
                "tempLowerThanMinutes": 0,
                "faultConditionExistsActive": False,
                "faultConditionExistsHours": 0,
                "normalConditionsActive": True,
                "communicationLostActive": False,
                "communicationLostHours": 0,
                "communicationFailureActive": True,
                "communicationFailureMinutes": 15,
                "deviceLostActive": False,
                "deviceLostHours": 0,
            },
            "isUpgrading": False,
            "isAlive": True,
            "thermostatVersion": "02.00.19.33",
            "macID": "00D02DEE4E56",
            "locationID": 2738909,
            "domainID": 20054,
            "instance": 0,
        },
        {
            "gatewayId": 2499896,
            "deviceID": 5333958,
            "thermostatModelType": "EMEA_ZONE",
            "deviceType": 128,
            "name": "Eh",
            "scheduleCapable": False,
            "holdUntilCapable": False,
            "thermostat": {
                "units": "Celsius",
                "indoorTemperature": 128.0,
                "outdoorTemperature": 128.0,
                "outdoorTemperatureAvailable": False,
                "outdoorHumidity": 128.0,
                "outdootHumidityAvailable": False,
                "indoorHumidity": 128.0,
                "indoorTemperatureStatus": "NotAvailable",
                "indoorHumidityStatus": "NotAvailable",
                "outdoorTemperatureStatus": "NotAvailable",
                "outdoorHumidityStatus": "NotAvailable",
                "isCommercial": False,
                "allowedModes": ["Heat", "Off"],
                "deadband": 0.0,
                "minHeatSetpoint": 5.0,
                "maxHeatSetpoint": 35.0,
                "minCoolSetpoint": 50.0,
                "maxCoolSetpoint": 90.0,
                "changeableValues": {
                    "mode": "Off",
                    "heatSetpoint": {"value": 21.0, "status": "Scheduled"},
                    "vacationHoldDays": 0,
                },
                "scheduleCapable": False,
                "vacationHoldChangeable": False,
                "vacationHoldCancelable": False,
                "scheduleHeatSp": 0.0,
                "scheduleCoolSp": 0.0,
            },
            "alertSettings": {
                "deviceID": 5333958,
                "tempHigherThanActive": True,
                "tempHigherThan": 30.0,
                "tempHigherThanMinutes": 0,
                "tempLowerThanActive": True,
                "tempLowerThan": 5.0,
                "tempLowerThanMinutes": 0,
                "faultConditionExistsActive": False,
                "faultConditionExistsHours": 0,
                "normalConditionsActive": True,
                "communicationLostActive": False,
                "communicationLostHours": 0,
                "communicationFailureActive": True,
                "communicationFailureMinutes": 15,
                "deviceLostActive": False,
                "deviceLostHours": 0,
            },
            "isUpgrading": False,
            "isAlive": True,
            "thermostatVersion": "02.00.19.33",
            "macID": "00D02DEE4E56",
            "locationID": 2738909,
            "domainID": 20054,
            "instance": 11,
        },
        {
            "gatewayId": 2499896,
            "deviceID": 3432577,
            "thermostatModelType": "EMEA_ZONE",
            "deviceType": 128,
            "name": "Front Room",
            "scheduleCapable": False,
            "holdUntilCapable": False,
            "thermostat": {
                "units": "Celsius",
                "indoorTemperature": 19.83,
                "outdoorTemperature": 128.0,
                "outdoorTemperatureAvailable": False,
                "outdoorHumidity": 128.0,
                "outdootHumidityAvailable": False,
                "indoorHumidity": 128.0,
                "indoorTemperatureStatus": "Measured",
                "indoorHumidityStatus": "NotAvailable",
                "outdoorTemperatureStatus": "NotAvailable",
                "outdoorHumidityStatus": "NotAvailable",
                "isCommercial": False,
                "allowedModes": ["Heat", "Off"],
                "deadband": 0.0,
                "minHeatSetpoint": 5.0,
                "maxHeatSetpoint": 35.0,
                "minCoolSetpoint": 50.0,
                "maxCoolSetpoint": 90.0,
                "changeableValues": {
                    "mode": "Off",
                    "heatSetpoint": {"value": 20.5, "status": "Scheduled"},
                    "vacationHoldDays": 0,
                },
                "scheduleCapable": False,
                "vacationHoldChangeable": False,
                "vacationHoldCancelable": False,
                "scheduleHeatSp": 0.0,
                "scheduleCoolSp": 0.0,
            },
            "alertSettings": {
                "deviceID": 3432577,
                "tempHigherThanActive": True,
                "tempHigherThan": 30.0,
                "tempHigherThanMinutes": 0,
                "tempLowerThanActive": True,
                "tempLowerThan": 5.0,
                "tempLowerThanMinutes": 0,
                "faultConditionExistsActive": False,
                "faultConditionExistsHours": 0,
                "normalConditionsActive": True,
                "communicationLostActive": False,
                "communicationLostHours": 0,
                "communicationFailureActive": True,
                "communicationFailureMinutes": 15,
                "deviceLostActive": False,
                "deviceLostHours": 0,
            },
            "isUpgrading": False,
            "isAlive": True,
            "thermostatVersion": "02.00.19.33",
            "macID": "00D02DEE4E56",
            "locationID": 2738909,
            "domainID": 20054,
            "instance": 2,
        },
        {
            "gatewayId": 2499896,
            "deviceID": 3449703,
            "thermostatModelType": "EMEA_ZONE",
            "deviceType": 128,
            "name": "Kids Room",
            "scheduleCapable": False,
            "holdUntilCapable": False,
            "thermostat": {
                "units": "Celsius",
                "indoorTemperature": 19.53,
                "outdoorTemperature": 128.0,
                "outdoorTemperatureAvailable": False,
                "outdoorHumidity": 128.0,
                "outdootHumidityAvailable": False,
                "indoorHumidity": 128.0,
                "indoorTemperatureStatus": "Measured",
                "indoorHumidityStatus": "NotAvailable",
                "outdoorTemperatureStatus": "NotAvailable",
                "outdoorHumidityStatus": "NotAvailable",
                "isCommercial": False,
                "allowedModes": ["Heat", "Off"],
                "deadband": 0.0,
                "minHeatSetpoint": 5.0,
                "maxHeatSetpoint": 35.0,
                "minCoolSetpoint": 50.0,
                "maxCoolSetpoint": 90.0,
                "changeableValues": {
                    "mode": "Off",
                    "heatSetpoint": {"value": 16.0, "status": "Scheduled"},
                    "vacationHoldDays": 0,
                },
                "scheduleCapable": False,
                "vacationHoldChangeable": False,
                "vacationHoldCancelable": False,
                "scheduleHeatSp": 0.0,
                "scheduleCoolSp": 0.0,
            },
            "alertSettings": {
                "deviceID": 3449703,
                "tempHigherThanActive": True,
                "tempHigherThan": 30.0,
                "tempHigherThanMinutes": 0,
                "tempLowerThanActive": True,
                "tempLowerThan": 5.0,
                "tempLowerThanMinutes": 0,
                "faultConditionExistsActive": False,
                "faultConditionExistsHours": 0,
                "normalConditionsActive": True,
                "communicationLostActive": False,
                "communicationLostHours": 0,
                "communicationFailureActive": True,
                "communicationFailureMinutes": 15,
                "deviceLostActive": False,
                "deviceLostHours": 0,
            },
            "isUpgrading": False,
            "isAlive": True,
            "thermostatVersion": "02.00.19.33",
            "macID": "00D02DEE4E56",
            "locationID": 2738909,
            "domainID": 20054,
            "instance": 6,
        },
        {
            "gatewayId": 2499896,
            "deviceID": 3432578,
            "thermostatModelType": "EMEA_ZONE",
            "deviceType": 128,
            "name": "Kitchen",
            "scheduleCapable": False,
            "holdUntilCapable": False,
            "thermostat": {
                "units": "Celsius",
                "indoorTemperature": 20.43,
                "outdoorTemperature": 128.0,
                "outdoorTemperatureAvailable": False,
                "outdoorHumidity": 128.0,
                "outdootHumidityAvailable": False,
                "indoorHumidity": 128.0,
                "indoorTemperatureStatus": "Measured",
                "indoorHumidityStatus": "NotAvailable",
                "outdoorTemperatureStatus": "NotAvailable",
                "outdoorHumidityStatus": "NotAvailable",
                "isCommercial": False,
                "allowedModes": ["Heat", "Off"],
                "deadband": 0.0,
                "minHeatSetpoint": 5.0,
                "maxHeatSetpoint": 35.0,
                "minCoolSetpoint": 50.0,
                "maxCoolSetpoint": 90.0,
                "changeableValues": {
                    "mode": "Off",
                    "heatSetpoint": {"value": 15.0, "status": "Scheduled"},
                    "vacationHoldDays": 0,
                },
                "scheduleCapable": False,
                "vacationHoldChangeable": False,
                "vacationHoldCancelable": False,
                "scheduleHeatSp": 0.0,
                "scheduleCoolSp": 0.0,
            },
            "alertSettings": {
                "deviceID": 3432578,
                "tempHigherThanActive": True,
                "tempHigherThan": 30.0,
                "tempHigherThanMinutes": 0,
                "tempLowerThanActive": True,
                "tempLowerThan": 5.0,
                "tempLowerThanMinutes": 0,
                "faultConditionExistsActive": False,
                "faultConditionExistsHours": 0,
                "normalConditionsActive": True,
                "communicationLostActive": False,
                "communicationLostHours": 0,
                "communicationFailureActive": True,
                "communicationFailureMinutes": 15,
                "deviceLostActive": False,
                "deviceLostHours": 0,
            },
            "isUpgrading": False,
            "isAlive": True,
            "thermostatVersion": "02.00.19.33",
            "macID": "00D02DEE4E56",
            "locationID": 2738909,
            "domainID": 20054,
            "instance": 3,
        },
        {
            "gatewayId": 2499896,
            "deviceID": 3432580,
            "thermostatModelType": "EMEA_ZONE",
            "deviceType": 128,
            "name": "Main Bedroom",
            "scheduleCapable": False,
            "holdUntilCapable": False,
            "thermostat": {
                "units": "Celsius",
                "indoorTemperature": 20.72,
                "outdoorTemperature": 128.0,
                "outdoorTemperatureAvailable": False,
                "outdoorHumidity": 128.0,
                "outdootHumidityAvailable": False,
                "indoorHumidity": 128.0,
                "indoorTemperatureStatus": "Measured",
                "indoorHumidityStatus": "NotAvailable",
                "outdoorTemperatureStatus": "NotAvailable",
                "outdoorHumidityStatus": "NotAvailable",
                "isCommercial": False,
                "allowedModes": ["Heat", "Off"],
                "deadband": 0.0,
                "minHeatSetpoint": 5.0,
                "maxHeatSetpoint": 35.0,
                "minCoolSetpoint": 50.0,
                "maxCoolSetpoint": 90.0,
                "changeableValues": {
                    "mode": "Off",
                    "heatSetpoint": {"value": 16.0, "status": "Scheduled"},
                    "vacationHoldDays": 0,
                },
                "scheduleCapable": False,
                "vacationHoldChangeable": False,
                "vacationHoldCancelable": False,
                "scheduleHeatSp": 0.0,
                "scheduleCoolSp": 0.0,
            },
            "alertSettings": {
                "deviceID": 3432580,
                "tempHigherThanActive": True,
                "tempHigherThan": 30.0,
                "tempHigherThanMinutes": 0,
                "tempLowerThanActive": True,
                "tempLowerThan": 5.0,
                "tempLowerThanMinutes": 0,
                "faultConditionExistsActive": False,
                "faultConditionExistsHours": 0,
                "normalConditionsActive": True,
                "communicationLostActive": False,
                "communicationLostHours": 0,
                "communicationFailureActive": True,
                "communicationFailureMinutes": 15,
                "deviceLostActive": False,
                "deviceLostHours": 0,
            },
            "isUpgrading": False,
            "isAlive": True,
            "thermostatVersion": "02.00.19.33",
            "macID": "00D02DEE4E56",
            "locationID": 2738909,
            "domainID": 20054,
            "instance": 5,
        },
        {
            "gatewayId": 2499896,
            "deviceID": 3432576,
            "thermostatModelType": "EMEA_ZONE",
            "deviceType": 128,
            "name": "Main Room",
            "scheduleCapable": False,
            "holdUntilCapable": False,
            "thermostat": {
                "units": "Celsius",
                "indoorTemperature": 20.14,
                "outdoorTemperature": 128.0,
                "outdoorTemperatureAvailable": False,
                "outdoorHumidity": 128.0,
                "outdootHumidityAvailable": False,
                "indoorHumidity": 128.0,
                "indoorTemperatureStatus": "Measured",
                "indoorHumidityStatus": "NotAvailable",
                "outdoorTemperatureStatus": "NotAvailable",
                "outdoorHumidityStatus": "NotAvailable",
                "isCommercial": False,
                "allowedModes": ["Heat", "Off"],
                "deadband": 0.0,
                "minHeatSetpoint": 5.0,
                "maxHeatSetpoint": 35.0,
                "minCoolSetpoint": 50.0,
                "maxCoolSetpoint": 90.0,
                "changeableValues": {
                    "mode": "Off",
                    "heatSetpoint": {"value": 15.0, "status": "Scheduled"},
                    "vacationHoldDays": 0,
                },
                "scheduleCapable": False,
                "vacationHoldChangeable": False,
                "vacationHoldCancelable": False,
                "scheduleHeatSp": 0.0,
                "scheduleCoolSp": 0.0,
            },
            "alertSettings": {
                "deviceID": 3432576,
                "tempHigherThanActive": True,
                "tempHigherThan": 30.0,
                "tempHigherThanMinutes": 0,
                "tempLowerThanActive": True,
                "tempLowerThan": 5.0,
                "tempLowerThanMinutes": 0,
                "faultConditionExistsActive": False,
                "faultConditionExistsHours": 0,
                "normalConditionsActive": True,
                "communicationLostActive": False,
                "communicationLostHours": 0,
                "communicationFailureActive": True,
                "communicationFailureMinutes": 15,
                "deviceLostActive": False,
                "deviceLostHours": 0,
            },
            "isUpgrading": False,
            "isAlive": True,
            "thermostatVersion": "02.00.19.33",
            "macID": "00D02DEE4E56",
            "locationID": 2738909,
            "domainID": 20054,
            "instance": 1,
        },
        {
            "gatewayId": 2499896,
            "deviceID": 3450733,
            "thermostatModelType": "EMEA_ZONE",
            "deviceType": 128,
            "name": "Spare Room",
            "scheduleCapable": False,
            "holdUntilCapable": False,
            "thermostat": {
                "units": "Celsius",
                "indoorTemperature": 18.81,
                "outdoorTemperature": 128.0,
                "outdoorTemperatureAvailable": False,
                "outdoorHumidity": 128.0,
                "outdootHumidityAvailable": False,
                "indoorHumidity": 128.0,
                "indoorTemperatureStatus": "Measured",
                "indoorHumidityStatus": "NotAvailable",
                "outdoorTemperatureStatus": "NotAvailable",
                "outdoorHumidityStatus": "NotAvailable",
                "isCommercial": False,
                "allowedModes": ["Heat", "Off"],
                "deadband": 0.0,
                "minHeatSetpoint": 5.0,
                "maxHeatSetpoint": 35.0,
                "minCoolSetpoint": 50.0,
                "maxCoolSetpoint": 90.0,
                "changeableValues": {
                    "mode": "Off",
                    "heatSetpoint": {"value": 16.0, "status": "Scheduled"},
                    "vacationHoldDays": 0,
                },
                "scheduleCapable": False,
                "vacationHoldChangeable": False,
                "vacationHoldCancelable": False,
                "scheduleHeatSp": 0.0,
                "scheduleCoolSp": 0.0,
            },
            "alertSettings": {
                "deviceID": 3450733,
                "tempHigherThanActive": True,
                "tempHigherThan": 30.0,
                "tempHigherThanMinutes": 0,
                "tempLowerThanActive": True,
                "tempLowerThan": 5.0,
                "tempLowerThanMinutes": 0,
                "faultConditionExistsActive": False,
                "faultConditionExistsHours": 0,
                "normalConditionsActive": True,
                "communicationLostActive": False,
                "communicationLostHours": 0,
                "communicationFailureActive": True,
                "communicationFailureMinutes": 15,
                "deviceLostActive": False,
                "deviceLostHours": 0,
            },
            "isUpgrading": False,
            "isAlive": True,
            "thermostatVersion": "02.00.19.33",
            "macID": "00D02DEE4E56",
            "locationID": 2738909,
            "domainID": 20054,
            "instance": 8,
        },
        {
            "gatewayId": 2499896,
            "deviceID": 5333957,
            "thermostatModelType": "EMEA_ZONE",
            "deviceType": 128,
            "name": "UFH",
            "scheduleCapable": False,
            "holdUntilCapable": False,
            "thermostat": {
                "units": "Celsius",
                "indoorTemperature": 128.0,
                "outdoorTemperature": 128.0,
                "outdoorTemperatureAvailable": False,
                "outdoorHumidity": 128.0,
                "outdootHumidityAvailable": False,
                "indoorHumidity": 128.0,
                "indoorTemperatureStatus": "NotAvailable",
                "indoorHumidityStatus": "NotAvailable",
                "outdoorTemperatureStatus": "NotAvailable",
                "outdoorHumidityStatus": "NotAvailable",
                "isCommercial": False,
                "allowedModes": ["Heat", "Off"],
                "deadband": 0.0,
                "minHeatSetpoint": 5.0,
                "maxHeatSetpoint": 35.0,
                "minCoolSetpoint": 50.0,
                "maxCoolSetpoint": 90.0,
                "changeableValues": {
                    "mode": "Off",
                    "heatSetpoint": {"value": 21.0, "status": "Scheduled"},
                    "vacationHoldDays": 0,
                },
                "scheduleCapable": False,
                "vacationHoldChangeable": False,
                "vacationHoldCancelable": False,
                "scheduleHeatSp": 0.0,
                "scheduleCoolSp": 0.0,
            },
            "alertSettings": {
                "deviceID": 5333957,
                "tempHigherThanActive": True,
                "tempHigherThan": 30.0,
                "tempHigherThanMinutes": 0,
                "tempLowerThanActive": True,
                "tempLowerThan": 5.0,
                "tempLowerThanMinutes": 0,
                "faultConditionExistsActive": False,
                "faultConditionExistsHours": 0,
                "normalConditionsActive": True,
                "communicationLostActive": False,
                "communicationLostHours": 0,
                "communicationFailureActive": True,
                "communicationFailureMinutes": 15,
                "deviceLostActive": False,
                "deviceLostHours": 0,
            },
            "isUpgrading": False,
            "isAlive": True,
            "thermostatVersion": "02.00.19.33",
            "macID": "00D02DEE4E56",
            "locationID": 2738909,
            "domainID": 20054,
            "instance": 10,
        },
        {
            "gatewayId": 2499896,
            "deviceID": 5333955,
            "thermostatModelType": "EMEA_ZONE",
            "deviceType": 128,
            "name": "Zv",
            "scheduleCapable": False,
            "holdUntilCapable": False,
            "thermostat": {
                "units": "Celsius",
                "indoorTemperature": 128.0,
                "outdoorTemperature": 128.0,
                "outdoorTemperatureAvailable": False,
                "outdoorHumidity": 128.0,
                "outdootHumidityAvailable": False,
                "indoorHumidity": 128.0,
                "indoorTemperatureStatus": "NotAvailable",
                "indoorHumidityStatus": "NotAvailable",
                "outdoorTemperatureStatus": "NotAvailable",
                "outdoorHumidityStatus": "NotAvailable",
                "isCommercial": False,
                "allowedModes": ["Heat", "Off"],
                "deadband": 0.0,
                "minHeatSetpoint": 5.0,
                "maxHeatSetpoint": 35.0,
                "minCoolSetpoint": 50.0,
                "maxCoolSetpoint": 90.0,
                "changeableValues": {
                    "mode": "Off",
                    "heatSetpoint": {"value": 21.0, "status": "Scheduled"},
                    "vacationHoldDays": 0,
                },
                "scheduleCapable": False,
                "vacationHoldChangeable": False,
                "vacationHoldCancelable": False,
                "scheduleHeatSp": 0.0,
                "scheduleCoolSp": 0.0,
            },
            "alertSettings": {
                "deviceID": 5333955,
                "tempHigherThanActive": True,
                "tempHigherThan": 30.0,
                "tempHigherThanMinutes": 0,
                "tempLowerThanActive": True,
                "tempLowerThan": 5.0,
                "tempLowerThanMinutes": 0,
                "faultConditionExistsActive": False,
                "faultConditionExistsHours": 0,
                "normalConditionsActive": True,
                "communicationLostActive": False,
                "communicationLostHours": 0,
                "communicationFailureActive": True,
                "communicationFailureMinutes": 15,
                "deviceLostActive": False,
                "deviceLostHours": 0,
            },
            "isUpgrading": False,
            "isAlive": True,
            "thermostatVersion": "02.00.19.33",
            "macID": "00D02DEE4E56",
            "locationID": 2738909,
            "domainID": 20054,
            "instance": 9,
        },
    ],
    "oneTouchButtons": [],
    "weather": {
        "condition": "NightClear",
        "temperature": 9.0,
        "units": "Celsius",
        "humidity": 87,
        "phrase": "Clear",
    },
    "daylightSavingTimeEnabled": True,
    "timeZone": {
        "id": "GMT Standard Time",
        "displayName": "(UTC+00:00) Dublin, Edinburgh, Lisbon, London",
        "offsetMinutes": 0,
        "currentOffsetMinutes": 0,
        "usingDaylightSavingTime": True,
    },
    "oneTouchActionsSuspended": False,
    "isLocationOwner": True,
    "locationOwnerID": 2263181,
    "locationOwnerName": "David Smith",
    "locationOwnerUserName": "null@gmail.com",
    "canSearchForContractors": True,
    "contractor": {
        "info": {"contractorID": 1839},
        "monitoring": {"levelOfAccess": "Partial", "contactPreferences": []},
    },
}
